@@ -1,4 +1,4 @@
-// Copyright (C) 2020 - 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2020 - 2023 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -22,7 +22,7 @@
 // This allows one to randomize the execution order for better a better experimental setup
 // which produces fewer type 1 errors where one incorrectly rejects the null hypothesis.
 
-#include <complex>
+#include <algorithm>
 #include <hip/hip_runtime_api.h>
 #include <iostream>
 #include <math.h>
@@ -38,7 +38,7 @@
 #endif
 
 #include "../../shared/gpubuf.h"
-#include "../rocfft_params.h"
+#include "../../shared/rocfft_params.h"
 #include "rider.h"
 #include "rocfft.h"
 
@@ -340,9 +340,12 @@ int main(int argc, char* argv[])
     // hip Device number for running tests:
     int deviceId{};
 
-    // Number of performance trial samples
+    // Number of performance trial samples:
     int ntrial{};
 
+    // Test sequence choice:
+    int test_sequence{};
+
     // Vector of test target libraries
     std::vector<std::string> libs;
 
@@ -362,8 +365,11 @@ int main(int argc, char* argv[])
         ("device", po::value<int>(&deviceId)->default_value(0), "Select a specific device id")
         ("verbose", po::value<int>(&verbose)->default_value(0), "Control output verbosity")
         ("ntrial,N", po::value<int>(&ntrial)->default_value(1), "Trial size for the problem")
+        ("sequence", po::value<int>(&test_sequence)->default_value(0),
+         "Test sequence: random(0), alternating(1) sequential(2)")
         ("notInPlace,o", "Not in-place FFT transform (default: in-place)")
-        ("double", "Double precision transform (default: single)")
+        ("double", "Double precision transform (deprecated: use --precision double)")
+        ("precision", po::value<fft_precision>(&params.precision), "Transform precision: single (default), double, half")
         ("transformType,t", po::value<fft_transform_type>(&params.transform_type)
          ->default_value(fft_transform_type_complex_forward),
          "Type of transform:\n0) complex forward\n1) complex inverse\n2) real "
@@ -394,7 +400,7 @@ int main(int argc, char* argv[])
         ("ioffset", po::value<std::vector<size_t>>(&params.ioffset)->multitoken(), "Input offsets.")
         ("ooffset", po::value<std::vector<size_t>>(&params.ooffset)->multitoken(), "Output offsets.")
         ("scalefactor", po::value<double>(&params.scale_factor), "Scale factor to apply to output.")
-        ("token", po::value<std::string>(&token));;
+        ("token", po::value<std::string>(&token));
     // clang-format on
 
     po::variables_map vm;
@@ -446,7 +452,8 @@ int main(int argc, char* argv[])
 
         params.placement
             = vm.count("notInPlace") ? fft_placement_notinplace : fft_placement_inplace;
-        params.precision = vm.count("double") ? fft_precision_double : fft_precision_single;
+        if(vm.count("double"))
+            params.precision = fft_precision_double;
 
         if(vm.count("notInPlace"))
         {
@@ -524,9 +531,14 @@ int main(int argc, char* argv[])
         std::cout << params.str() << std::endl;
     }
 
+    // Check free and total available memory:
+    size_t free  = 0;
+    size_t total = 0;
+    HIP_V_THROW(hipMemGetInfo(&free, &total), "hipMemGetInfo failed");
+
     const auto raw_vram_footprint
         = params.fft_params_vram_footprint() + twiddle_table_vram_footprint(params);
-    if(!vram_fits_problem(raw_vram_footprint))
+    if(!vram_fits_problem(raw_vram_footprint, free))
     {
         std::cout << "SKIPPED: Problem size (" << raw_vram_footprint
                   << ") raw data too large for device.\n";
@@ -534,7 +546,7 @@ int main(int argc, char* argv[])
     }
 
     const auto vram_footprint = params.vram_footprint();
-    if(!vram_fits_problem(vram_footprint))
+    if(!vram_fits_problem(vram_footprint, free))
     {
         std::cout << "SKIPPED: Problem size (" << vram_footprint
                   << ") raw data too large for device.\n";
@@ -618,7 +630,7 @@ int main(int argc, char* argv[])
     }
 
     // Input data:
-    compute_input(params, ibuffer);
+    params.compute_input(ibuffer);
 
     if(verbose > 1)
     {
@@ -671,21 +683,64 @@ int main(int argc, char* argv[])
     // Execution times for loaded libraries:
     std::vector<std::vector<double>> time(libs.size());
 
+    std::vector<int> testcase(ntrial * libs.size());
+    switch(test_sequence)
+    {
+    case 0:
+    {
+        // Random order:
+        for(int itrial = 0; itrial < ntrial; ++itrial)
+        {
+            for(size_t ilib = 0; ilib < libs.size(); ++ilib)
+            {
+                testcase[libs.size() * itrial + ilib] = ilib;
+            }
+        }
+
+        std::random_device rd;
+        std::mt19937       g(rd());
+        std::shuffle(testcase.begin(), testcase.end(), g);
+        break;
+    }
+    case 1:
+        // Alternating order:
+        for(int itrial = 0; itrial < ntrial; ++itrial)
+        {
+            for(size_t ilib = 0; ilib < libs.size(); ++ilib)
+            {
+                testcase[libs.size() * itrial + ilib] = ilib;
+            }
+        }
+        break;
+    case 2:
+        // Sequential order:
+        for(int itrial = 0; itrial < ntrial; ++itrial)
+        {
+            for(size_t ilib = 0; ilib < libs.size(); ++ilib)
+            {
+                testcase[ilib * ntrial + itrial] = ilib;
+            }
+        }
+
+        break;
+    default:
+        throw std::runtime_error("Invalid test sequence choice.");
+    }
+
+    std::cout << "test case:";
+    for(const auto i : testcase)
+        std::cout << " " << i;
+    std::cout << "\n";
+
     // Run the FFTs from the different libraries in random order until they all have at
     // least ntrial times.
     std::vector<int> ndone(libs.size());
     std::fill(ndone.begin(), ndone.end(), 0);
-    while(!std::all_of(ndone.begin(), ndone.end(), [&ntrial](int i) { return (i >= ntrial); }))
+    for(size_t itest = 0; itest < testcase.size(); ++itest)
     {
-        const int idx = rand() % ndone.size();
-        ndone[idx]++;
+        const int idx = testcase[itest];
 
-        // We can optionally require that all runs have exactly ntrial, but it may be more
-        // iid to just let things run:
-        // if(ndone[idx] > ntrial)
-        //     continue;
-
-        compute_input(params, ibuffer);
+        params.compute_input(ibuffer);
 
         // Run the plan using its associated rocFFT library:
         time[idx].push_back(
