@@ -1,4 +1,4 @@
-// Copyright (C) 2016 - 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2016 - 2023 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -34,7 +34,6 @@
 #include "../../shared/concurrency.h"
 #include "../../shared/environment.h"
 #include "../../shared/work_queue.h"
-#include "../rocfft_params.h"
 #include "rocfft.h"
 #include "rocfft_accuracy_test.h"
 #include "test_params.h"
@@ -53,14 +52,27 @@ int verbose;
 
 // User-defined random seed
 size_t random_seed;
+// Probability of running individual planar FFTs
+double planar_prob;
+// Probability of running individual callback FFTs
+double callback_prob;
 
 // Transform parameters for manual test:
 fft_params manual_params;
 
-// Ram limitation for tests (GiB):
+// Host memory limitation for tests (GiB):
 size_t ramgb;
 
+// Device memory limitation for tests (GiB):
+size_t vramgb;
+
+// Allow skipping tests if there is a runtime error
+bool skip_runtime_fails;
+// But count the number of failures
+int n_hip_failures = 0;
+
 // Manually specified precision cutoffs:
+double half_epsilon;
 double single_epsilon;
 double double_epsilon;
 
@@ -69,6 +81,8 @@ double max_linf_eps_double = 0.0;
 double max_l2_eps_double   = 0.0;
 double max_linf_eps_single = 0.0;
 double max_l2_eps_single   = 0.0;
+double max_linf_eps_half   = 0.0;
+double max_l2_eps_half     = 0.0;
 
 // Control whether we use FFTW's wisdom (which we use to imply FFTW_MEASURE).
 bool use_fftw_wisdom = false;
@@ -76,27 +90,31 @@ bool use_fftw_wisdom = false;
 // Cache the last cpu fft that was requested
 last_cpu_fft_cache last_cpu_fft_data;
 
-static size_t get_system_memory_GiB()
+system_memory get_system_memory()
 {
-    // system memory often has a little chunk carved out for other
-    // stuff, so round up to nearest GiB.
+    system_memory memory_data;
 #ifdef WIN32
     MEMORYSTATUSEX info;
     info.dwLength = sizeof(info);
     if(!GlobalMemoryStatusEx(&info))
-        return 0;
-    return (info.ullTotalPhys + ONE_GiB - 1) / ONE_GiB;
+        return memory_data;
+    memory_data.total_bytes = info.ullTotalPhys;
+    memory_data.free_bytes  = info.ullAvailPhys;
 #else
     struct sysinfo info;
     if(sysinfo(&info) != 0)
-        return 0;
-    return (info.totalram * info.mem_unit + ONE_GiB - 1) / ONE_GiB;
+        return memory_data;
+    memory_data.total_bytes = info.totalram * info.mem_unit;
+    memory_data.free_bytes  = info.freeram * info.mem_unit;
 #endif
+    return memory_data;
 }
 
+system_memory start_memory = get_system_memory();
+
 void precompile_test_kernels(const std::string& precompile_file)
 {
-    std::cout << "precompiling test kernels..." << std::endl;
+    std::cout << "precompiling test kernels...\n";
     WorkQueue<std::string> tokenQueue;
 
     std::vector<std::string> tokens;
@@ -137,7 +155,7 @@ void precompile_test_kernels(const std::string& precompile_file)
     std::mt19937       dist(dev());
     std::shuffle(tokens.begin(), tokens.end(), dist);
     auto precompile_begin = std::chrono::steady_clock::now();
-    std::cout << "precompiling " << tokens.size() << " FFT plans..." << std::endl;
+    std::cout << "precompiling " << tokens.size() << " FFT plans...\n";
 
     for(auto&& t : tokens)
         tokenQueue.push(std::move(t));
@@ -153,10 +171,30 @@ void precompile_test_kernels(const std::string& precompile_file)
                 std::string token{tokenQueue.pop()};
                 if(token.empty())
                     break;
-                rocfft_params params;
-                params.from_token(token);
-                params.validate();
-                params.setup_structs();
+
+                try
+                {
+                    rocfft_params params_forward;
+                    params_forward.from_token(token);
+                    params_forward.validate();
+                    params_forward.setup_structs();
+
+                    params_forward.free();
+
+                    rocfft_params params_inverse;
+                    params_inverse.inverse_from_forward(params_forward);
+                    params_inverse.validate();
+                    params_inverse.setup_structs();
+                }
+                catch(std::exception& e)
+                {
+                    // failed to create a plan, abort
+                    //
+                    // we could continue on, but the test should just
+                    // fail later anyway in the same way.  so report
+                    // which token failed early and get out
+                    throw std::runtime_error(token + " plan creation failure: " + e.what());
+                }
             }
         });
         // insert empty tokens to tell threads to stop
@@ -168,25 +206,12 @@ void precompile_test_kernels(const std::string& precompile_file)
     auto                                      precompile_end = std::chrono::steady_clock::now();
     std::chrono::duration<double, std::milli> precompile_ms  = precompile_end - precompile_begin;
     std::cout << "done precompiling FFT plans in " << static_cast<size_t>(precompile_ms.count())
-              << " ms" << std::endl;
+              << " ms\n";
 }
 
 int main(int argc, char* argv[])
 {
-    // NB: If we initialize gtest first, then it removes all of its own command-line
-    // arguments and sets argc and argv correctly; no need to jump through hoops for
-    // boost::program_options.
-    ::testing::InitGoogleTest(&argc, argv);
-
-    // Filename for fftw and fftwf wisdom.
-    std::string fftw_wisdom_filename;
-
-    // Token string to fully specify fft params for the manual test.
-    std::string test_token;
-
-    // Filename for precompiled kernels to be written to
-    std::string precompile_file;
-
+    // We would like to parse a few arguments before initiating gtest.
     po::options_description opdesc(
         "\n"
         "rocFFT Runtime Test command line options\n"
@@ -206,13 +231,46 @@ int main(int argc, char* argv[])
         "      HP - hermitian planar\n"
         "\n"
         "Usage");
+    // clang-format off
+    opdesc.add_options()
+        ("verbose,v",
+         po::value<int>()->default_value(0),
+         "print out detailed information for the tests.")
+        ("seed", po::value<size_t>(&random_seed),
+         "Random seed; if unset, use an actual random seed.")
+        ("planar_prob", po::value<double>(&planar_prob)->default_value(0.1),
+        "Probability of running individual planar transforms")
+        ("callback_prob", po::value<double>(&callback_prob)->default_value(0.1),
+         "Probability of running individual callback transforms");
+    // clang-format on
+    po::variables_map vm;
+    po::store(po::command_line_parser(argc, argv).options(opdesc).allow_unregistered().run(), vm);
+    po::notify(vm);
+
+    verbose = vm["verbose"].as<int>();
+
+    // NB: If we initialize gtest first, then it removes all of its own command-line
+    // arguments and sets argc and argv correctly; no need to jump through hoops for
+    // boost::program_options.
+    ::testing::InitGoogleTest(&argc, argv);
+
+    // Filename for fftw and fftwf wisdom.
+    std::string fftw_wisdom_filename;
+
+    // Token string to fully specify fft params for the manual test.
+    std::string test_token;
+
+    // Filename for precompiled kernels to be written to
+    std::string precompile_file;
+
     // Declare the supported options.
     // clang-format doesn't handle boost program options very well:
     // clang-format off
     opdesc.add_options()
         ("help,h", "produces this help message")
-        ("verbose,v",  po::value<int>()->default_value(0),
-        "print out detailed information for the tests.")
+        ("skip_runtime_fails",  po::value<bool>(&skip_runtime_fails)->default_value(true),
+        "Skip the test if there is a runtime failure.")
+        ("version", "Print queryable version information from the rocfft library and exit")
         ("transformType,t", po::value<fft_transform_type>(&manual_params.transform_type)
          ->default_value(fft_transform_type_complex_forward),
          "Type of transform:\n0) complex forward\n1) complex inverse\n2) real "
@@ -220,7 +278,9 @@ int main(int argc, char* argv[])
         ("notInPlace,o", "Not in-place FFT transform (default: in-place)")
         ("callback", "Inject load/store callbacks")
         ("checkstride", "Check that data is not written outside of output strides")
-        ("double", "Double precision transform (default: single)")
+        ("double", "Double precision transform (deprecated: use --precision double)")
+        ("precision", po::value<fft_precision>(&manual_params.precision),
+         "Transform precision: single (default), double, half")
         ( "itype", po::value<fft_array_type>(&manual_params.itype)
           ->default_value(fft_array_type_unset),
           "Array type of input data:\n0) interleaved\n1) planar\n2) real\n3) "
@@ -248,39 +308,46 @@ int main(int argc, char* argv[])
          "Logical size of input buffer.")
         ("osize", po::value<std::vector<size_t>>(&manual_params.osize)->multitoken(),
          "Logical size of output.")
-        ("R", po::value<size_t>(&ramgb)->default_value(get_system_memory_GiB()), "Ram limit in GiB for tests.")
-        ("single_epsilon",  po::value<double>(&single_epsilon)->default_value(3.75e-5)) 
-	("double_epsilon",  po::value<double>(&double_epsilon)->default_value(1e-15))
+        ("R", po::value<size_t>(&ramgb)->default_value((start_memory.total_bytes + ONE_GiB - 1) / ONE_GiB), "Ram limit in GiB for tests.")
+        ("V", po::value<size_t>(&vramgb)->default_value(0), "vram limit in GiB for tests.")
+        ("half_epsilon",  po::value<double>(&half_epsilon)->default_value(9.77e-4))
+        ("single_epsilon",  po::value<double>(&single_epsilon)->default_value(3.75e-5))
+        ("double_epsilon",  po::value<double>(&double_epsilon)->default_value(1e-15))
         ("wise,w", "use FFTW wisdom")
         ("wisdomfile,W",
          po::value<std::string>(&fftw_wisdom_filename)->default_value("wisdom3.txt"),
          "FFTW3 wisdom filename")
         ("scalefactor", po::value<double>(&manual_params.scale_factor), "Scale factor to apply to output.")
         ("token", po::value<std::string>(&test_token)->default_value(""), "Test token name for manual test")
-        ("precompile",  po::value<std::string>(&precompile_file), "Precompile kernels to a file for all test cases before running tests")
-        ("seed", po::value<size_t>(&random_seed), "Random seed; if unset, use an actual random seed.");
+        ("precompile",  po::value<std::string>(&precompile_file), "Precompile kernels to a file for all test cases before running tests");
     // clang-format on
 
-    po::variables_map vm;
     po::store(po::parse_command_line(argc, argv, opdesc), vm);
     po::notify(vm);
 
     if(vm.count("help"))
     {
-        std::cout << opdesc << std::endl;
+        std::cout << opdesc << "\n";
         return 0;
     }
-    verbose = vm["verbose"].as<int>();
 
-    std::cout << "single epsilon: " << single_epsilon << "\tdouble epsilon: " << double_epsilon
-              << std::endl;
+    if(vm.count("version"))
+    {
+        char v[256];
+        rocfft_get_version_string(v, 256);
+        std::cout << "version " << v << "\n";
+        return EXIT_SUCCESS;
+    }
+
+    std::cout << "half epsilon: " << half_epsilon << "\tsingle epsilon: " << single_epsilon
+              << "\tdouble epsilon: " << double_epsilon << "\n";
 
     if(!vm.count("seed"))
     {
         std::random_device dev;
         random_seed = dev();
     }
-    std::cout << "Random seed: " << random_seed << std::endl;
+    std::cout << "Random seed: " << random_seed << "\n";
 
     if(vm.count("wise"))
     {
@@ -302,7 +369,7 @@ int main(int argc, char* argv[])
     rocfft_setup();
     char v[256];
     rocfft_get_version_string(v, 256);
-    std::cout << "rocFFT version: " << v << std::endl;
+    std::cout << "rocFFT version: " << v << "\n";
 
 #ifdef FFTW_MULTITHREAD
     fftw_init_threads();
@@ -359,7 +426,7 @@ int main(int argc, char* argv[])
 
     if(test_token != "")
     {
-        std::cout << "Reading fft params from token:\n" << test_token << std::endl;
+        std::cout << "Reading fft params from token:\n" << test_token << "\n";
 
         try
         {
@@ -367,7 +434,7 @@ int main(int argc, char* argv[])
         }
         catch(...)
         {
-            std::cout << "Unable to parse token." << std::endl;
+            std::cout << "Unable to parse token.\n";
             return 1;
         }
     }
@@ -381,7 +448,8 @@ int main(int argc, char* argv[])
 
         manual_params.placement
             = vm.count("notInPlace") ? fft_placement_notinplace : fft_placement_inplace;
-        manual_params.precision = vm.count("double") ? fft_precision_double : fft_precision_single;
+        if(vm.count("double"))
+            manual_params.precision = fft_precision_double;
 
         if(vm.count("callback"))
         {
@@ -424,23 +492,27 @@ int main(int argc, char* argv[])
 
     rocfft_cleanup();
 
-    std::cout << "single precision max l-inf epsilon: " << max_linf_eps_single << std::endl;
-    std::cout << "single precision max l2 epsilon:     " << max_l2_eps_single << std::endl;
-    std::cout << "double precision max l-inf epsilon: " << max_linf_eps_double << std::endl;
-    std::cout << "double precision max l2 epsilon:     " << max_l2_eps_double << std::endl;
+    std::cout << "Random seed: " << random_seed << "\n";
+    std::cout << "half precision max l-inf epsilon: " << max_linf_eps_half << "\n";
+    std::cout << "half precision max l2 epsilon:     " << max_l2_eps_half << "\n";
+    std::cout << "single precision max l-inf epsilon: " << max_linf_eps_single << "\n";
+    std::cout << "single precision max l2 epsilon:     " << max_l2_eps_single << "\n";
+    std::cout << "double precision max l-inf epsilon: " << max_linf_eps_double << "\n";
+    std::cout << "double precision max l2 epsilon:     " << max_l2_eps_double << "\n";
+    std::cout << "Number of runtime issues: " << n_hip_failures << "\n";
 
     return retval;
 }
 
-TEST(manual, vs_fftw)
+TEST(manual, vs_fftw) // MANUAL TESTS HERE
 {
     // Run an individual test using the provided command-line parameters.
     manual_params.validate();
 
     std::cout << "Manual test:"
-              << "\n\t" << manual_params.str("\n\t") << std::endl;
+              << "\n\t" << manual_params.str("\n\t") << "\n";
 
-    std::cout << "Token: " << manual_params.token() << std::endl;
+    std::cout << "Token: " << manual_params.token() << "\n";
 
     if(!manual_params.valid(verbose + 2))
     {
