@@ -1,5 +1,5 @@
 /******************************************************************************
-* Copyright (C) 2016 - 2022 Advanced Micro Devices, Inc. All rights reserved.
+* Copyright (C) 2016 - 2023 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
@@ -24,8 +24,8 @@
  *  rocfft.h defines all the public interfaces and types
  *  */
 
-#ifndef __ROCFFT_H__
-#define __ROCFFT_H__
+#ifndef ROCFFT_H
+#define ROCFFT_H
 
 #ifdef rocfft_EXPORTS
 #include "rocfft-export.h"
@@ -84,6 +84,7 @@ typedef enum rocfft_precision_e
 {
     rocfft_precision_single,
     rocfft_precision_double,
+    rocfft_precision_half,
 } rocfft_precision;
 
 /*! @brief Result placement
@@ -213,7 +214,7 @@ ROCFFT_EXPORT rocfft_status rocfft_plan_destroy(rocfft_plan plan);
  *  @details rocFFT multiplies each element of the result by the given factor at the end of the transform.
  *
  *  The supplied factor must be a finite number.  That is, it must neither be infinity nor NaN.
- * 
+ *
  *  @param[in] description description handle
  *  @param[in] scale scaling factor
  *  */
@@ -222,10 +223,10 @@ ROCFFT_EXPORT rocfft_status rocfft_plan_description_set_scale_factor(
 
 /*!
  *  @brief Set advanced data layout parameters on a plan description
- * 
+ *
  *  @details This API specifies advanced layout of input/output
  *  buffers for a plan description.
- * 
+ *
  *  The following parameters are supported for inputs and outputs:
  *
  *  * Array type (real, hermitian, or complex data, in either
@@ -508,8 +509,18 @@ ROCFFT_EXPORT rocfft_status rocfft_cache_buffer_free(void* buffer);
 ROCFFT_EXPORT rocfft_status rocfft_cache_deserialize(const void* buffer, size_t buffer_len_bytes);
 #endif
 
+#ifdef ROCFFT_BUILD_OFFLINE_TUNER
+/*! @brief Get a handler of offline-tuner
+
+ *  @details This is for developers only, so the actual type is not
+ *  public to API yet. This API is used only in our standalone executable.
+ *  Must be called after rocfft_setup. And de-references the return handle to get
+ *  the tuner-pointer */
+ROCFFT_EXPORT rocfft_status rocfft_get_offline_tuner_handle(void** offline_tuner);
+#endif
+
 #ifdef __cplusplus
 }
 #endif /* __cplusplus */
 
-#endif /* __ROCFFT_H__ */
+#endif /* ROCFFT_H */
