@@ -792,14 +792,21 @@ class Map(BaseNodeOps):
     def emplace(self, key, value):
         return Call(self.name + '.emplace', arguments=ArgumentList(key, value))
 
-    def assert_emplace(self, key, value):
+    def assert_emplace(self, key, value, what_error):
         emplace = Call(self.name + '.emplace',
                        arguments=ArgumentList(key, value)).inline()
         status = Call(name='std::get<1>',
                       arguments=ArgumentList(emplace)).inline()
-        throw = StatementList(Throw('std::runtime_error("' + str(key) + '")'))
+        throw = StatementList(
+            Throw('std::runtime_error("' + str(what_error) + '")'))
         return If(Equal(status, "false"), throw)
 
+    def assert_insert(self, key, value):
+        insert = Call('insert_default_entry',
+                      arguments=ArgumentList(key, value)).inline()
+        throw = StatementList(Throw('std::runtime_error("' + str(key) + '")'))
+        return If(Equal(insert, "false"), throw)
+
     # def __getitem__(self, idx):
     #     return ArrayElement(self.name, idx)
 
@@ -879,20 +886,6 @@ class Using(BaseNode):
         return f'using {self.name} = {self.spec};'
 
 
-@name_args(['name', 'arguments', 'templates', 'qualifier'])
-class Prototype(BaseNode):
-
-    def __str__(self) -> str:
-        f = ''
-        if self.templates:
-            f += 'template<' + str(self.templates) + '>'
-        if self.qualifier is not None:
-            f += self.qualifier + ' '
-        f += ' void ' + self.name
-        f += '(' + str(self.arguments) + ') FUNCTION_POOL_STANDALONE_BODY ;'
-        return f
-
-
 @name_args([
     'name', 'value', 'arguments', 'templates', 'qualifier', 'launch_bounds',
     'body', 'meta'
@@ -916,10 +909,6 @@ class Function(BaseNode):
         f += '{' + njoin(self.body) + '}'
         return f
 
-    def prototype(self):
-        return Prototype(self.name, self.arguments, self.templates,
-                         self.qualifier)
-
     def address(self):
         return Address(self.name)
 
