@@ -21,8 +21,8 @@
 #pragma once
 
 #include <algorithm>
-#include <any>
 #include <iostream>
+#include <memory>
 #include <numeric>
 #include <optional>
 #include <string.h>
@@ -77,7 +77,6 @@ enum class Component
 // Expressions
 //
 
-struct ScalarVariable;
 class Variable;
 class Literal;
 class ComplexLiteral;
@@ -111,6 +110,7 @@ class Ternary;
 // FFT expressions
 
 class LoadGlobal;
+class LoadGlobalPlanar;
 
 class TwiddleMultiply;
 class TwiddleMultiplyConjugate;
@@ -120,9 +120,9 @@ class Parens;
 class CallExpr;
 
 class IntrinsicLoad;
+class IntrinsicLoadPlanar;
 
-using Expression = std::variant<ScalarVariable,
-                                Variable,
+using Expression = std::variant<Variable,
                                 Literal,
                                 ComplexLiteral,
                                 Add,
@@ -148,22 +148,24 @@ using Expression = std::variant<ScalarVariable,
                                 PreDecrement,
                                 Ternary,
                                 LoadGlobal,
+                                LoadGlobalPlanar,
                                 TwiddleMultiply,
                                 TwiddleMultiplyConjugate,
                                 Parens,
                                 CallExpr,
-                                IntrinsicLoad>;
+                                IntrinsicLoad,
+                                IntrinsicLoadPlanar>;
 
 class OptionalExpression
 {
-    std::any expr;
+    std::unique_ptr<Expression> expr;
 
 public:
-    OptionalExpression(){};
-    OptionalExpression(OptionalExpression&&)      = default;
-    OptionalExpression(const OptionalExpression&) = default;
-    OptionalExpression& operator=(OptionalExpression&&) = default;
-    OptionalExpression& operator=(const OptionalExpression&) = default;
+    OptionalExpression();
+    OptionalExpression(OptionalExpression&&);
+    OptionalExpression(const OptionalExpression&);
+    OptionalExpression& operator=(OptionalExpression&&);
+    OptionalExpression& operator=(const OptionalExpression&);
     explicit OptionalExpression(const Expression& expr);
     OptionalExpression& operator=(const Expression& in_expr);
     Expression          operator*() const;
@@ -206,35 +208,13 @@ public:
     }
 };
 
-struct ScalarVariable
-{
-    static const unsigned int precedence = 0;
-    std::string               name, type;
-    Component                 component;
-    OptionalExpression        index;
-
-    ScalarVariable(const std::string& name,
-                   const std::string& type,
-                   Component          component = Component::BOTH)
-        : name(name)
-        , type(type)
-        , component(component){};
-    ScalarVariable(ScalarVariable&&)      = default;
-    ScalarVariable(const ScalarVariable&) = default;
-    ScalarVariable& operator=(const ScalarVariable&) = default;
-    ScalarVariable& operator=(ScalarVariable&&) = default;
-
-    std::string render() const;
-};
-
 class Variable
 {
 public:
     static const unsigned int precedence = 0;
     std::string               name, type;
     bool                      pointer = false, restrict = false;
-    ScalarVariable            x, y;
-    Component                 component;
+    Component                 component = Component::BOTH;
     // index2d + size2d are set if this is a 2D array variable
     OptionalExpression index;
     OptionalExpression index2D;
@@ -255,12 +235,6 @@ public:
              bool restrict,
              const Expression& _size);
 
-    Variable(const ScalarVariable& v)
-        : name(v.name)
-        , type(v.type)
-        , x(v.name + ".x", v.type)
-        , y(v.name + ".y", v.type){};
-
     Variable(const Variable& v);
     Variable(Variable&& v) = default;
     Variable(const Variable& v, const Expression& _index);
@@ -270,8 +244,12 @@ public:
     Variable& operator=(Variable&&) = default;
     Variable  operator[](const Expression& index) const;
     // do a 2D array access
-    Variable       at(const Expression& index, const Expression& index2D) const;
-    ScalarVariable address() const;
+    Variable at(const Expression& index, const Expression& index2D) const;
+    Variable address() const;
+
+    // assuming this is a complex value, access the x, y members
+    Variable x() const;
+    Variable y() const;
 
     std::string render() const;
 };
@@ -384,22 +362,35 @@ public:
     std::vector<Expression> args;
 };
 
+class LoadGlobalPlanar
+{
+public:
+    static const unsigned int precedence = 18;
+    explicit LoadGlobalPlanar(const std::vector<Expression>& args);
+    LoadGlobalPlanar(LoadGlobalPlanar&&)      = default;
+    LoadGlobalPlanar(const LoadGlobalPlanar&) = default;
+    LoadGlobalPlanar& operator=(LoadGlobalPlanar&&) = default;
+    LoadGlobalPlanar& operator=(const LoadGlobalPlanar&) = default;
+
+    std::string render() const;
+
+    std::vector<Expression> args;
+};
+
 class TwiddleMultiply
 {
 public:
     static const unsigned int precedence = 18;
     TwiddleMultiply(const Variable& a, const Variable& b)
-        : a(a)
-        , b(b)
+        : vars({a, b})
     {
     }
     TwiddleMultiply(TwiddleMultiply&&)      = default;
     TwiddleMultiply(const TwiddleMultiply&) = default;
-    TwiddleMultiply& operator=(TwiddleMultiply&&) = default;
-    TwiddleMultiply& operator=(const TwiddleMultiply&) = default;
-    Variable         a;
-    Variable         b;
-    std::string      render() const;
+    TwiddleMultiply&      operator=(TwiddleMultiply&&) = default;
+    TwiddleMultiply&      operator=(const TwiddleMultiply&) = default;
+    std::vector<Variable> vars;
+    std::string           render() const;
 };
 
 class TwiddleMultiplyConjugate
@@ -407,16 +398,14 @@ class TwiddleMultiplyConjugate
 public:
     static const unsigned int precedence = 18;
     TwiddleMultiplyConjugate(const Variable& a, const Variable& b)
-        : a(a)
-        , b(b)
+        : vars({a, b})
     {
     }
     TwiddleMultiplyConjugate(TwiddleMultiplyConjugate&&)      = default;
     TwiddleMultiplyConjugate(const TwiddleMultiplyConjugate&) = default;
     TwiddleMultiplyConjugate& operator=(TwiddleMultiplyConjugate&&) = default;
     TwiddleMultiplyConjugate& operator=(const TwiddleMultiplyConjugate&) = default;
-    Variable                  a;
-    Variable                  b;
+    std::vector<Variable>     vars;
     std::string               render() const;
 };
 
@@ -452,10 +441,25 @@ public:
     std::string             render() const;
 };
 
+class IntrinsicLoadPlanar
+{
+public:
+    static const unsigned int precedence = 18;
+    explicit IntrinsicLoadPlanar(const std::vector<Expression>& args);
+    IntrinsicLoadPlanar(IntrinsicLoadPlanar&&)      = default;
+    IntrinsicLoadPlanar(const IntrinsicLoadPlanar&) = default;
+    IntrinsicLoadPlanar& operator=(IntrinsicLoadPlanar&&) = default;
+    IntrinsicLoadPlanar& operator=(const IntrinsicLoadPlanar&) = default;
+
+    // data, voffset, soffset, rw
+    std::vector<Expression> args;
+    std::string             render() const;
+};
+
 #define MAKE_OPER(NAME, OPER, PRECEDENCE)                           \
     class NAME                                                      \
     {                                                               \
-        std::string oper{OPER};                                     \
+        static constexpr const char* oper = {OPER};                 \
                                                                     \
     public:                                                         \
         static const unsigned int precedence = PRECEDENCE;          \
@@ -604,8 +608,11 @@ static PreDecrement operator--(const Expression& a)
 // Expressions, for some classes.
 
 class Assign;
+class ReturnExpr;
 class Call;
 class CallbackDeclaration;
+class CallbackLoadDeclaration;
+class CallbackStoreDeclaration;
 class Declaration;
 class LDSDeclaration;
 class For;
@@ -618,6 +625,7 @@ class StoreGlobalPlanar;
 class StatementList;
 class Butterfly;
 class IntrinsicStore;
+class IntrinsicStorePlanar;
 class IntrinsicLoadToDest;
 
 struct LineBreak
@@ -674,8 +682,11 @@ struct CommentLines
 };
 
 using Statement = std::variant<Assign,
+                               ReturnExpr,
                                Call,
                                CallbackDeclaration,
+                               CallbackLoadDeclaration,
+                               CallbackStoreDeclaration,
                                CommentLines,
                                Declaration,
                                LDSDeclaration,
@@ -692,6 +703,7 @@ using Statement = std::variant<Assign,
                                SyncThreads,
                                Butterfly,
                                IntrinsicStore,
+                               IntrinsicStorePlanar,
                                IntrinsicLoadToDest>;
 
 class Assign
@@ -792,6 +804,49 @@ public:
     }
 };
 
+class CallbackLoadDeclaration
+{
+public:
+    CallbackLoadDeclaration(const std::string& scalar_type, const std::string& cbtype)
+        : scalar_type(scalar_type)
+        , cbtype(cbtype){};
+    std::string scalar_type;
+    std::string cbtype;
+    std::string render() const
+    {
+        return "auto load_cb = get_load_cb<" + scalar_type + ", " + cbtype + ">(load_cb_fn);";
+    }
+};
+
+class CallbackStoreDeclaration
+{
+public:
+    CallbackStoreDeclaration(const std::string& scalar_type, const std::string& cbtype)
+        : scalar_type(scalar_type)
+        , cbtype(cbtype){};
+    std::string scalar_type;
+    std::string cbtype;
+    std::string render() const
+    {
+        return "auto store_cb = get_store_cb<" + scalar_type + ", " + cbtype + ">(store_cb_fn);";
+    }
+};
+
+class ReturnExpr
+{
+public:
+    ReturnExpr(const Expression& expr)
+        : expr(expr)
+    {
+    }
+
+    Expression expr;
+
+    std::string render() const
+    {
+        return "return " + vrender(expr) + ";";
+    }
+};
 class Call
 {
 public:
@@ -919,12 +974,12 @@ public:
     {
         // Output two assignments
         return Assign{realPtr[index],
-                      scale_factor ? Expression{value.x * scale_factor.value()}
-                                   : Expression{value.x}}
+                      scale_factor ? Expression{value.x() * scale_factor.value()}
+                                   : Expression{value.x()}}
                    .render()
                + Assign{imagPtr[index],
-                        scale_factor ? Expression{value.y * scale_factor.value()}
-                                     : Expression{value.y}}
+                        scale_factor ? Expression{value.y() * scale_factor.value()}
+                                     : Expression{value.y()}}
                      .render();
     }
 
@@ -979,6 +1034,52 @@ public:
     std::optional<Expression> scale_factor;
 };
 
+class IntrinsicStorePlanar
+{
+public:
+    IntrinsicStorePlanar(const Expression&                ptrre,
+                         const Expression&                ptrim,
+                         const Expression&                voffset,
+                         const Expression&                soffset,
+                         const Expression&                value,
+                         const Expression&                rw_flag,
+                         const std::optional<Expression>& scale_factor)
+        : ptrre{ptrre}
+        , ptrim{ptrim}
+        , voffset{voffset}
+        , soffset{soffset}
+        , value{value}
+        , rw_flag{rw_flag}
+        , scale_factor{scale_factor}
+    {
+    }
+    std::string render() const
+    {
+        return "store_intrinsic(" + vrender(ptrre) + "," + vrender(voffset) + "," + vrender(soffset)
+               + ","
+               + Literal{"real_type_t<scalar_type>("
+                         + vrender(scale_factor ? (value * scale_factor.value()) : value) + ".x)"}
+                     .render()
+               + "," + vrender(rw_flag)
+               + ");"
+                 "\n"
+                 "store_intrinsic("
+               + vrender(ptrim) + "," + vrender(voffset) + "," + vrender(soffset) + ","
+               + Literal{"real_type_t<scalar_type>("
+                         + vrender(scale_factor ? (value * scale_factor.value()) : value) + ".y)"}
+                     .render()
+               + "," + vrender(rw_flag) + ");";
+    }
+
+    Expression                ptrre;
+    Expression                ptrim;
+    Expression                voffset;
+    Expression                soffset;
+    Expression                value;
+    Expression                rw_flag;
+    std::optional<Expression> scale_factor;
+};
+
 class IntrinsicLoadToDest
 {
 public:
@@ -1049,6 +1150,7 @@ public:
     ArgumentList  arguments;
     TemplateList  templates;
     std::string   qualifier;
+    std::string   return_type   = "void";
     unsigned int  launch_bounds = 0;
 
     explicit Function(const std::string& name)
@@ -1078,7 +1180,6 @@ struct BaseVisitor
         return visit_##CLS(x);          \
     }
 
-    MAKE_VISITOR_OPERATOR(Expression, ScalarVariable);
     MAKE_VISITOR_OPERATOR(Expression, Variable);
     MAKE_VISITOR_OPERATOR(Expression, Literal);
     MAKE_VISITOR_OPERATOR(Expression, ComplexLiteral);
@@ -1104,16 +1205,21 @@ struct BaseVisitor
     MAKE_VISITOR_OPERATOR(Expression, PreDecrement);
     MAKE_VISITOR_OPERATOR(Expression, Ternary);
     MAKE_VISITOR_OPERATOR(Expression, LoadGlobal);
+    MAKE_VISITOR_OPERATOR(Expression, LoadGlobalPlanar);
     MAKE_VISITOR_OPERATOR(Expression, ComplexMultiply);
     MAKE_VISITOR_OPERATOR(Expression, TwiddleMultiply);
     MAKE_VISITOR_OPERATOR(Expression, TwiddleMultiplyConjugate);
     MAKE_VISITOR_OPERATOR(Expression, Parens);
     MAKE_VISITOR_OPERATOR(Expression, CallExpr);
     MAKE_VISITOR_OPERATOR(Expression, IntrinsicLoad);
+    MAKE_VISITOR_OPERATOR(Expression, IntrinsicLoadPlanar);
 
     MAKE_VISITOR_OPERATOR(StatementList, Assign);
+    MAKE_VISITOR_OPERATOR(StatementList, ReturnExpr);
     MAKE_VISITOR_OPERATOR(StatementList, Call);
     MAKE_VISITOR_OPERATOR(StatementList, CallbackDeclaration);
+    MAKE_VISITOR_OPERATOR(StatementList, CallbackLoadDeclaration);
+    MAKE_VISITOR_OPERATOR(StatementList, CallbackStoreDeclaration);
     MAKE_VISITOR_OPERATOR(StatementList, CommentLines);
     MAKE_VISITOR_OPERATOR(StatementList, Declaration);
     MAKE_VISITOR_OPERATOR(StatementList, LDSDeclaration);
@@ -1130,6 +1236,7 @@ struct BaseVisitor
     MAKE_VISITOR_OPERATOR(StatementList, SyncThreads);
     MAKE_VISITOR_OPERATOR(StatementList, Butterfly);
     MAKE_VISITOR_OPERATOR(StatementList, IntrinsicStore);
+    MAKE_VISITOR_OPERATOR(StatementList, IntrinsicStorePlanar);
     MAKE_VISITOR_OPERATOR(StatementList, IntrinsicLoadToDest);
 
     MAKE_VISITOR_OPERATOR(ArgumentList, ArgumentList);
@@ -1207,19 +1314,23 @@ struct BaseVisitor
     MAKE_EXPR_VISIT(PreDecrement);
 
     MAKE_EXPR_VISIT(LoadGlobal);
+    MAKE_EXPR_VISIT(LoadGlobalPlanar);
 
     MAKE_TRIVIAL_VISIT(Expression, ComplexMultiply);
     MAKE_TRIVIAL_VISIT(Expression, TwiddleMultiply);
     MAKE_TRIVIAL_VISIT(Expression, TwiddleMultiplyConjugate);
 
     MAKE_EXPR_VISIT(IntrinsicLoad);
+    MAKE_EXPR_VISIT(IntrinsicLoadPlanar);
     MAKE_EXPR_VISIT(Parens);
 
     MAKE_EXPR_VISIT(Ternary);
     MAKE_EXPR_VISIT(ComplexLiteral)
 
-    MAKE_TRIVIAL_VISIT(Expression, ScalarVariable)
+    MAKE_TRIVIAL_STATEMENT_VISIT(ReturnExpr)
     MAKE_TRIVIAL_STATEMENT_VISIT(CallbackDeclaration)
+    MAKE_TRIVIAL_STATEMENT_VISIT(CallbackLoadDeclaration)
+    MAKE_TRIVIAL_STATEMENT_VISIT(CallbackStoreDeclaration)
     MAKE_TRIVIAL_STATEMENT_VISIT(LDSDeclaration)
 
     MAKE_TRIVIAL_VISIT(Expression, Literal)
@@ -1342,6 +1453,21 @@ struct BaseVisitor
         return StatementList{IntrinsicStore(ptr, voffset, soffset, value, rw_flag)};
     }
 
+    virtual StatementList visit_IntrinsicStorePlanar(const IntrinsicStorePlanar& x)
+    {
+        auto                      ptrre   = std::visit(*this, x.ptrre);
+        auto                      ptrim   = std::visit(*this, x.ptrim);
+        auto                      voffset = std::visit(*this, x.voffset);
+        auto                      soffset = std::visit(*this, x.soffset);
+        auto                      value   = std::visit(*this, x.value);
+        auto                      rw_flag = std::visit(*this, x.rw_flag);
+        std::optional<Expression> scale_factor;
+        if(x.scale_factor)
+            scale_factor = std::visit(*this, x.scale_factor.value());
+        return StatementList{
+            IntrinsicStorePlanar(ptrre, ptrim, voffset, soffset, value, rw_flag, scale_factor)};
+    }
+
     virtual StatementList visit_StoreGlobalPlanar(const StoreGlobalPlanar& x)
     {
         auto                      realPtr = std::get<Variable>(visit_Variable(x.realPtr));
@@ -1454,7 +1580,7 @@ struct MakePlanarVisitor : public BaseVisitor
                 auto im = ptr;
                 im.name = imname;
 
-                stmts += Assign{x.lhs, ComplexLiteral{re[idx], im[idx]}, x.oper};
+                stmts += Assign{x.lhs, LoadGlobalPlanar({re, im, idx}), x.oper};
                 return stmts;
             }
         }
@@ -1474,10 +1600,8 @@ struct MakePlanarVisitor : public BaseVisitor
                 auto im = ptr;
                 im.name = imname;
 
-                stmts += Assign{x.lhs,
-                                ComplexLiteral{IntrinsicLoad({re, voffset, soffset, rw_flag}),
-                                               IntrinsicLoad({im, voffset, soffset, rw_flag})},
-                                x.oper};
+                stmts += Assign{
+                    x.lhs, IntrinsicLoadPlanar({re, im, voffset, soffset, rw_flag}), x.oper};
                 return stmts;
             }
         }
@@ -1516,26 +1640,9 @@ struct MakePlanarVisitor : public BaseVisitor
             auto im = var;
             im.name = imname;
 
-            StatementList stmts;
-            stmts += Call{
-                "store_intrinsic",
-                {re,
-                 x.voffset,
-                 x.soffset,
-                 Literal{"real_type_t<scalar_type>("
-                         + vrender(x.scale_factor ? (x.scale_factor.value() * x.value) : x.value)
-                         + ".x)"},
-                 x.rw_flag}};
-            stmts += Call{
-                "store_intrinsic",
-                {im,
-                 x.voffset,
-                 x.soffset,
-                 Literal{"real_type_t<scalar_type>("
-                         + vrender(x.scale_factor ? (x.scale_factor.value() * x.value) : x.value)
-                         + ".y)"},
-                 x.rw_flag}};
-            return stmts;
+            auto value = std::get<Variable>(x.value);
+            return {IntrinsicStorePlanar{
+                re, im, x.voffset, x.soffset, value, x.rw_flag, x.scale_factor}};
         }
         return StatementList{x};
     }
@@ -1726,7 +1833,7 @@ struct MakeInverseVisitor : public BaseVisitor
 {
     Expression visit_TwiddleMultiply(const TwiddleMultiply& x) override
     {
-        return TwiddleMultiplyConjugate{x.a, x.b};
+        return TwiddleMultiplyConjugate{x.vars[0], x.vars[1]};
     }
 
     StatementList visit_Butterfly(const Butterfly& x) override
