@@ -43,7 +43,7 @@ from operator import mul
 
 from generator import (ArgumentList, BaseNode, Call, CommentBlock, Function,
                        Include, LineBreak, Map, StatementList, Variable,
-                       name_args, write, clang_format_file)
+                       Assign, name_args, write)
 
 from collections import namedtuple
 
@@ -81,13 +81,6 @@ def cjoin(xs):
 #
 # Helpers
 #
-
-
-def flatten(lst):
-    """Flatten a list of lists to a list."""
-    return sum(lst, [])
-
-
 def unique(kernels):
     """Merge kernel lists without duplicated meta.length; ignore later ones."""
     r, s = list(), set()
@@ -123,8 +116,10 @@ class FFTKernel(BaseNode):
             f += str(self.function.address())
         use_3steps_large_twd = getattr(self.function.meta,
                                        'use_3steps_large_twd', None)
+        # assume half-precision needs the same thing as single
+        precision = 'sp' if self.function.meta.precision == 'half' else self.function.meta.precision
         if use_3steps_large_twd is not None:
-            f += ', ' + str(use_3steps_large_twd[self.function.meta.precision])
+            f += ', ' + str(use_3steps_large_twd[precision])
         else:
             f += ', false'
         factors = getattr(self.function.meta, 'factors', None)
@@ -164,24 +159,27 @@ def generate_cpu_function_pool(functions):
     function_map = Map('function_map')
     precisions = {
         'sp': 'rocfft_precision_single',
-        'dp': 'rocfft_precision_double'
+        'dp': 'rocfft_precision_double',
+        'half': 'rocfft_precision_half',
     }
+    var_kernel = Variable('kernel', 'FFTKernel')
 
     populate = StatementList()
+    populate += var_kernel.declaration()
     for f in functions:
         length, precision, scheme, transpose = f.meta.length, f.meta.precision, f.meta.scheme, f.meta.transpose
         if isinstance(length, (int, str)):
             length = [length, 0]
+        populate += Assign(var_kernel, FFTKernel(f))
         key = Call(name='std::make_tuple',
                    arguments=ArgumentList(
                        'std::array<size_t, 2>({' + cjoin(length) + '})',
-                       precisions[precision], scheme, transpose
-                       or 'NONE')).inline()
-        populate += function_map.assert_emplace(key, FFTKernel(f))
+                       precisions[precision], scheme, transpose or 'NONE',
+                       'kernel.get_kernel_config()')).inline()
+        populate += function_map.assert_insert(key, var_kernel)
 
     return StatementList(
         Include('"../include/function_pool.h"'),
-        StatementList(*[f.prototype() for f in functions]),
         Function(name='function_pool::function_pool',
                  value=False,
                  arguments=ArgumentList(),
@@ -642,7 +640,9 @@ def list_large_kernels():
         if not hasattr(k, 'length'):
             k.length = functools.reduce(lambda a, b: a * b, k.factors)
 
-    # SBRC
+    # for SBRC, if direct_to_from_reg is True, we do store-from-reg, but will not do load-to-reg
+    #           And since SBRC is is dir-from-lds but NOT dir-to-reg, the global load part requires full LDS
+    #           So, SBRC is able to use half-lds.
     sbrc_kernels = [
         NS(length=17,  factors=[17], scheme='CS_KERNEL_STOCKHAM_BLOCK_RC', workgroup_size=256, threads_per_transform=1, runtime_compile=True),
         NS(length=49,  factors=[7, 7], scheme='CS_KERNEL_STOCKHAM_BLOCK_RC', workgroup_size=196, threads_per_transform=7), # block_width=28
@@ -668,6 +668,9 @@ def list_large_kernels():
         NS(length=1331, factors=[11, 11, 11], scheme='CS_KERNEL_STOCKHAM_BLOCK_RC', workgroup_size=256, threads_per_transform=121, runtime_compile=True),
     ]
 
+    for k in sbrc_kernels:
+        k.half_lds = False
+
     # NB:
     # Technically, we could have SBCR kernels the same amount as SBCC.
     #
@@ -677,8 +680,8 @@ def list_large_kernels():
     #
 
     # for SBCR, if direct_to_from_reg is True, we do load-to-reg, but will not do store-from-reg
-    # TODO- tune on ROCm 5.2, SBCR 56 shows no improvement while it has on ROCm 5.0
-    #       Need to be aware of any compiler optimization / regression
+    #           And since sbcr is dir-to-reg BUT NOT dir-from-reg, the global store part requires full LDS
+    #           So, we can't satifly half_lds in SBCR !
     sbcr_kernels = [
         NS(length=56,  factors=[7, 8], direct_to_from_reg=False),
         NS(length=100, factors=[10, 10], workgroup_size=100),
@@ -689,11 +692,10 @@ def list_large_kernels():
     block_width = 16
     for k in sbcr_kernels:
         k.scheme = 'CS_KERNEL_STOCKHAM_BLOCK_CR'
+        k.half_lds = False
         if not hasattr(k, 'workgroup_size'):
             k.workgroup_size = block_width * \
                 functools.reduce(mul, k.factors, 1) // min(k.factors)
-        if hasattr(k, 'half_lds') and k.half_lds is True:
-            k.workgroup_size = min(1024, k.workgroup_size * 2)
         if not hasattr(k, 'length'):
             k.length = functools.reduce(lambda a, b: a * b, k.factors)
 
@@ -744,6 +746,10 @@ def generate_kernel(kernel, precisions, stockham_aot):
     # default half_lds to True only for CS_KERNEL_STOCKHAM
     half_lds = getattr(kernel, 'half_lds',
                        kernel.scheme == 'CS_KERNEL_STOCKHAM')
+    # but we don't use LDS for single-radix kernels, so half_lds is meaningless there
+    if len(kernel.factors) == 1:
+        half_lds = False
+
     # for unspecified direct_to_from_reg, default is True only for CS_KERNEL_STOCKHAM and SBCC
     direct_to_from_reg = getattr(kernel, 'direct_to_from_reg', True)
 
@@ -755,16 +761,15 @@ def generate_kernel(kernel, precisions, stockham_aot):
     args.append(kernel.scheme)
     args.append(filename)
 
-    proc = subprocess.Popen(args=args, stderr=subprocess.PIPE)
+    proc = subprocess.Popen(args=args)
     ret_code = proc.wait()
     if (ret_code != 0):
-        print(proc.stderr.read().decode('ascii'))
         sys.exit(f"Error executing " + stockham_aot)
 
     kernel_metadata_file = open(kernel_file_name(kernel) + '.json', 'r')
     launchers = json.load(kernel_metadata_file)
 
-    clang_format_file(filename)
+    # don't format generated source files since they aren't currently used
 
     cpu_functions = []
     data = Variable('data_p', 'const void *')
@@ -797,23 +802,27 @@ def generate_kernel(kernel, precisions, stockham_aot):
             threads_per_transform, 0
         ]
 
-        f = Function(name=launcher.name,
-                     arguments=ArgumentList(data, back),
-                     meta=NS(
-                         factors=factors,
-                         length=length,
-                         params=params,
-                         precision=precision,
-                         runtime_compile=runtime_compile,
-                         scheme=scheme,
-                         workgroup_size=workgroup_size,
-                         transforms_per_block=transforms_per_block,
-                         threads_per_transform=tpt_list,
-                         transpose=sbrc_transpose_type,
-                         use_3steps_large_twd=use_3steps_large_twd,
-                     ))
+        precisions = [precision]
+        if precision == 'sp':
+            precisions.append('half')
+        for p in precisions:
+            f = Function(name=launcher.name,
+                         arguments=ArgumentList(data, back),
+                         meta=NS(
+                             factors=factors,
+                             length=length,
+                             params=params,
+                             precision=p,
+                             runtime_compile=runtime_compile,
+                             scheme=scheme,
+                             workgroup_size=workgroup_size,
+                             transforms_per_block=transforms_per_block,
+                             threads_per_transform=tpt_list,
+                             transpose=sbrc_transpose_type,
+                             use_3steps_large_twd=use_3steps_large_twd,
+                         ))
 
-        cpu_functions.append(f)
+            cpu_functions.append(f)
 
     return cpu_functions
 
@@ -826,52 +835,12 @@ def generate_kernels(kernels, precisions, stockham_aot):
 
     A list of CPU functions is returned.
     """
-    import concurrent.futures
-    import queue
 
-    # push all the work to a queue
-    q_in = queue.Queue()
+    ret = []
     for k in kernels:
-        q_in.put(k)
+        ret += generate_kernel(k, precisions, stockham_aot)
 
-    # queue for outputs
-    q_out = queue.Queue()
-
-    def threadfunc():
-        nonlocal q_in
-        nonlocal q_out
-        nonlocal precisions
-        nonlocal stockham_aot
-        try:
-            while not q_in.empty():
-                k = q_in.get()
-                q_out.put(generate_kernel(k, precisions, stockham_aot))
-        except queue.Empty:
-            pass
-
-    # by default, start up worker threads.  disable this if you want
-    # to use pdb to debug
-    use_threads = True
-
-    if use_threads:
-        with concurrent.futures.ThreadPoolExecutor() as executor:
-            results = []
-            for i in range(os.cpu_count()):
-                results.append(executor.submit(threadfunc))
-            for result in results:
-                result.result()
-    else:
-        threadfunc()
-
-    # iterate over the queue
-    def queue_iter(q_out):
-        try:
-            while not q_out.empty():
-                yield q_out.get()
-        except queue.Empty:
-            pass
-
-    return flatten(queue_iter(q_out))
+    return ret
 
 
 def cli():
