@@ -1,4 +1,4 @@
-// Copyright (C) 2016 - 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2016 - 2023 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -30,23 +30,16 @@
 #include <vector>
 
 #include "../../../shared/gpubuf.h"
+#include "../../../shared/rocfft_complex.h"
 #include "../device/kernels/callback.h"
 #include "../device/kernels/common.h"
 #include "compute_scheme.h"
+#include "enum_printer.h"
+#include "function_map_key.h"
 #include "kargs.h"
 #include "rtc_kernel.h"
 #include <hip/hip_runtime_api.h>
 
-enum OperatingBuffer
-{
-    OB_UNINIT              = 0b00000,
-    OB_USER_IN             = 0b00001,
-    OB_USER_OUT            = 0b00010,
-    OB_TEMP                = 0b00100,
-    OB_TEMP_CMPLX_FOR_REAL = 0b01000,
-    OB_TEMP_BLUESTEIN      = 0b10000,
-};
-
 enum NodeType
 {
     NT_UNDEFINED, // un init
@@ -65,20 +58,6 @@ enum FuseType
     FT_STOCKHAM_R2C_TRANSPOSE, // Stokham + post-r2c + transpose (Advance of FT_R2C_TRANSPOSE)
 };
 
-// TODO: move this to rocfft.h and allow users to select via plan description
-// the decision strategy for buffer assigment
-enum rocfft_optimize_strategy
-{
-    rocfft_optimize_min_buffer, // minimize number of buffers, possibly fewer fusions
-    rocfft_optimize_balance, // balance between buffer and fusion
-    rocfft_optimize_max_fusion, // maximize number of fusions, possibly more buffers
-};
-
-std::string PrintOperatingBuffer(const OperatingBuffer ob);
-std::string PrintOperatingBufferCode(const OperatingBuffer ob);
-std::string PrintSBRCTransposeType(const SBRC_TRANSPOSE_TYPE ty);
-std::string PrintDirectToFromRegMode(const DirectRegType ty);
-
 typedef void (*DevFnCall)(const void*, void*);
 
 struct GridParam
@@ -100,6 +79,47 @@ struct GridParam
     }
 };
 
+// get device property
+static hipDeviceProp_t get_curr_device_prop()
+{
+    hipDeviceProp_t prop;
+    int             deviceId = 0;
+    if(hipGetDevice(&deviceId) != hipSuccess)
+        throw std::runtime_error("hipGetDevice failed.");
+
+    if(hipGetDeviceProperties(&prop, deviceId) != hipSuccess)
+        throw std::runtime_error("hipGetDeviceProperties failed for deviceId "
+                                 + std::to_string(deviceId));
+
+    return prop;
+}
+
+// get the arch name, as a part of key of solution map
+static std::string get_arch_name(const hipDeviceProp_t& prop)
+{
+    static const std::vector<std::string> arch_list = {"gfx803",
+                                                       "gfx900",
+                                                       "gfx906",
+                                                       "gfx908",
+                                                       "gfx90a",
+                                                       "gfx1030",
+                                                       "gfx1100",
+                                                       "gfx1101",
+                                                       "gfx1102"};
+
+    static const std::string anyArch("any");
+    std::string              archName(prop.gcnArchName);
+
+    for(const auto& arch : arch_list)
+    {
+        if(archName.find(arch) != std::string::npos)
+            return arch;
+    }
+
+    // kind of a fall-back solution
+    return anyArch;
+}
+
 static bool is_device_gcn_arch(const hipDeviceProp_t& prop, const std::string& cmpTarget)
 {
     std::string archName(prop.gcnArchName);
@@ -117,19 +137,37 @@ static bool is_cube_size(const std::vector<size_t>& length)
     return length.size() == 3 && length[0] == length[1] && length[1] == length[2];
 }
 
-inline size_t sizeof_precision(rocfft_precision precision)
+// Given a map of precision-length exceptions, check whether the
+// length is present.  Assume half-precision has the same exceptions
+// as single-precision.
+static bool length_excepted(const std::map<rocfft_precision, std::set<size_t>>& exceptions,
+                            rocfft_precision                                    precision,
+                            size_t                                              length)
 {
-    switch(precision)
-    {
-    case rocfft_precision_single:
-        return 2 * sizeof(float);
-    case rocfft_precision_double:
-        return 2 * sizeof(double);
-    }
-    assert(false);
-    return 0;
+    if(precision == rocfft_precision_half)
+        precision = rocfft_precision_single;
+    return exceptions.at(precision).count(length);
 }
 
+void get_large_twd_base_steps(size_t large1DLen, bool use3steps, size_t& base, size_t& steps);
+
+struct SchemeTree
+{
+    ComputeScheme                            curScheme;
+    size_t                                   numKernels = 0;
+    std::vector<std::unique_ptr<SchemeTree>> children;
+
+    SchemeTree() {}
+    SchemeTree(ComputeScheme s)
+        : curScheme(s)
+    {
+    }
+};
+
+using SchemeVec = std::vector<ComputeScheme>;
+
+static SchemeVec EmptySchemeVec = {};
+
 class TreeNode;
 
 // The mininal tree node data needed to decide the scheme
@@ -140,7 +178,9 @@ struct NodeMetaData
     std::vector<size_t>     length;
     std::vector<size_t>     outputLength;
     std::vector<size_t>     inStride, outStride;
+    std::vector<size_t>     inStrideBlue, outStrideBlue;
     size_t                  iDist = 0, oDist = 0;
+    size_t                  iDistBlue = 0, oDistBlue = 0;
     size_t                  iOffset = 0, oOffset = 0;
     int                     direction    = -1;
     rocfft_result_placement placement    = rocfft_placement_inplace;
@@ -266,9 +306,15 @@ public:
     // Stride of the FFT in each dimension
     std::vector<size_t> inStride, outStride;
 
+    // Stride of the fused Bluestein FFT in each dimension
+    std::vector<size_t> inStrideBlue, outStrideBlue;
+
     // Distance between consecutive batch members:
     size_t iDist = 0, oDist = 0;
 
+    // Distance between consecutive batch members in fused Bluestein nodes
+    size_t iDistBlue = 0, oDistBlue = 0;
+
     // Offsets to start of data in buffer:
     size_t iOffset = 0, oOffset = 0;
 
@@ -294,7 +340,7 @@ public:
     bool largeTwd3Steps = false;
     // "Steps": how many exact loops we need to decompose the LTWD?
     // if we pass this as a template arg in kernel, should avoid dynamic while-loop
-    // We will update this in set_large_twd_base_steps()
+    // We will update this in get_large_twd_base_steps()
     size_t ltwdSteps = 0;
     // true if large twd multiply uses batch as transform count - this
     // is done on strided large 1D FFTs where the batch dimension moves
@@ -308,7 +354,10 @@ public:
     DirectRegType dir2regMode = DirectRegType::FORCE_OFF_OR_NOT_SUPPORT;
 
     // sbrc transpose type
-    SBRC_TRANSPOSE_TYPE sbrcTranstype = SBRC_TRANSPOSE_TYPE::NONE;
+    mutable SBRC_TRANSPOSE_TYPE sbrcTranstype = SBRC_TRANSPOSE_TYPE::NONE;
+
+    // specified kernel key from solution map. (if there is any)
+    std::unique_ptr<FMKey> specified_key;
 
     // Tree structure:
     // non-owning pointer to parent node, may be null
@@ -326,7 +375,13 @@ public:
     // Length of the FFT for computing zero-padded linear convolutions
     // in Bluestein's algorithm. If Bluestein is required to compute an
     // FFT of length N, then lengthBlue >= 2N - 1.
-    size_t lengthBlue = 0;
+    size_t lengthBlue  = 0;
+    size_t lengthBlueN = 0;
+
+    //
+    BluesteinType     typeBlue   = BluesteinType::BT_NONE;
+    BluesteinFuseType fuseBlue   = BluesteinFuseType::BFT_NONE;
+    bool              need_chirp = false;
 
     // Device pointers:
     // twiddle memory is owned by the repo
@@ -334,6 +389,8 @@ public:
     size_t           twiddles_size       = 0;
     void*            twiddles_large      = nullptr;
     size_t           twiddles_large_size = 0;
+    void*            chirp               = nullptr;
+    size_t           chirp_size          = 0;
     gpubuf_t<size_t> devKernArg;
 
     // callback parameters
@@ -404,8 +461,10 @@ public:
         return false;
     }
 
-    virtual void RecursiveBuildTree(); // Main tree builder: override by child
-    virtual void SanityCheck();
+    void RecursiveBuildTree(SchemeTree* solution_scheme = nullptr);
+
+    virtual void SanityCheck(SchemeTree*         solution_scheme = nullptr,
+                             std::vector<FMKey>& kernel_keys     = EmptyFMKeyVec);
     // If high dims are contiguous, we can collapse them to make offset
     // calculation simpler
     void CollapseContiguousDims();
@@ -440,11 +499,14 @@ public:
                                size_t& chirpSize);
 
     // Output plan information for debug purposes:
-    void Print(rocfft_ostream& os, int indent = 0) const;
+    virtual void Print(rocfft_ostream& os, int indent = 0) const;
 
     // logic B - using in-place transposes, todo
     //void RecursiveBuildTreeLogicB();
 
+    void RecursiveFindChildNodes(const ComputeScheme& scheme, std::vector<TreeNode*>& nodes);
+    void RecursiveCopyNodeData(const TreeNode& srcNode);
+
     void RecursiveRemoveNode(TreeNode* node);
 
     // insert a newNode before the node "pos"
@@ -474,10 +536,10 @@ public:
         return false;
     }
 
-    virtual bool KernelCheck()                                             = 0;
-    virtual bool CreateDevKernelArgs()                                     = 0;
-    virtual bool CreateTwiddleTableResource()                              = 0;
-    virtual void SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp) = 0;
+    virtual bool KernelCheck(std::vector<FMKey>& kernel_keys = EmptyFMKeyVec) = 0;
+    virtual bool CreateDevKernelArgs()                                        = 0;
+    virtual bool CreateDeviceResources()                                      = 0;
+    virtual void SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp)    = 0;
 
     // for 3D SBRC kernels, decide the transpose type based on the
     // block width and lengths that the block tiles need to align on.
@@ -487,6 +549,16 @@ public:
         return NONE;
     }
 
+    // default implementation of leaf node, for non-sbrc type without sbrc_trans
+    virtual FMKey GetKernelKey() const
+    {
+        if(specified_key)
+            return *specified_key.get();
+
+        return (dimension == 1) ? fpkey(length[0], precision, scheme)
+                                : fpkey(length[0], length[1], precision, scheme);
+    }
+
     // Compute the large twd decomposition base
     void set_large_twd_base_steps(size_t largeTWDLength);
 
@@ -496,8 +568,8 @@ public:
     bool IsBluesteinChirpSetup();
 
 protected:
-    virtual void BuildTree_internal()    = 0;
-    virtual void AssignParams_internal() = 0;
+    virtual void BuildTree_internal(const SchemeVec& child_schemes = EmptySchemeVec) = 0;
+    virtual void AssignParams_internal()                                             = 0;
 };
 
 class InternalNode : public TreeNode
@@ -517,9 +589,9 @@ protected:
         return false;
     }
 
-    bool CreateTwiddleTableResource() override
+    bool CreateDeviceResources() override
     {
-        throw std::runtime_error("Shouldn't call CreateTwiddleTableResource in a non-LeafNode");
+        throw std::runtime_error("Shouldn't call CreateDeviceResources in a non-LeafNode");
         return false;
     }
 
@@ -529,7 +601,7 @@ protected:
     }
 
 public:
-    bool KernelCheck() override
+    bool KernelCheck(std::vector<FMKey>& kernel_keys = EmptyFMKeyVec) override
     {
         return true;
     }
@@ -557,9 +629,11 @@ public:
     size_t              wgs              = 0;
     size_t              lds              = 0;
 
-    void           BuildTree_internal() final {} // nothing to do in leaf node
-    void           AssignParams_internal() final {} // nothing to do in leaf node
-    bool           CreateLargeTwdTable();
+    void BuildTree_internal(const SchemeVec& child_schemes = EmptySchemeVec) final {
+    } // nothing to do in leaf node
+    void AssignParams_internal() final {} // nothing to do in leaf node
+    bool CreateLargeTwdTable();
+
     virtual size_t GetTwiddleTableLength();
     // Limit length of generated twiddle table.  Default limit is 0,
     // which means to generate the full length of table.
@@ -569,11 +643,16 @@ public:
     }
     virtual void SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp) = 0;
 
-    bool         KernelCheck() override;
-    void         SanityCheck() override;
+public:
+    // leaf node would print additional informations about kernel setting
+    void         Print(rocfft_ostream& os, int indent = 0) const override;
+    bool         KernelCheck(std::vector<FMKey>& kernel_keys = EmptyFMKeyVec) override;
+    void         SanityCheck(SchemeTree*         solution_scheme = nullptr,
+                             std::vector<FMKey>& kernel_keys     = EmptyFMKeyVec) override;
     virtual bool CreateDevKernelArgs() override;
-    bool         CreateTwiddleTableResource() override;
+    bool         CreateDeviceResources() override;
     void         SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp) override;
+    FMKey        GetKernelKey() const override;
     virtual void GetKernelFactors();
 };
 
@@ -620,6 +699,12 @@ struct ExecPlan
     // are the nodes that do actual work
     std::vector<TreeNode*> execSeq;
 
+    // kernels that extracted from solution map
+    std::vector<FMKey> solution_kernels;
+
+    // scheme decompositions from solution map
+    std::unique_ptr<SchemeTree> rootScheme;
+
     // flattened potentially-fusable shims of rootPlan
     std::vector<FuseShim*> fuseShims;
 
@@ -631,6 +716,12 @@ struct ExecPlan
     std::vector<size_t> iLength;
     std::vector<size_t> oLength;
 
+    // Indicates whether this is a standalone chirp plan
+    // in multi-kernel Bluestein implementations (buffers
+    // in the standalone plan are not connected with the
+    // rest of the nodes in the fft plan).
+    bool IsChirpPlan;
+
     // default: starting from ABT, balance buffers and fusions
     // we could allow users to set in the later PR
     rocfft_optimize_strategy assignOptStrategy = rocfft_optimize_balance;
@@ -654,6 +745,10 @@ struct ExecPlan
     std::pair<TreeNode*, TreeNode*> get_load_store_nodes() const;
 };
 
+std::unique_ptr<SchemeTree> ApplySolution(ExecPlan& execPlan);
+
+// get a min_token (without batch, stride, offset...) of a node, for generating a prob-key
+void GetNodeToken(const TreeNode& probNode, std::string& min_token, std::string& full_token);
 void ProcessNode(ExecPlan& execPlan);
 void PrintNode(rocfft_ostream& os, const ExecPlan& execPlan);
 
