@@ -1,4 +1,4 @@
-// Copyright (C) 2021 - 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2021 - 2023 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -20,6 +20,7 @@
 
 #include "node_factory.h"
 #include "../../shared/arithmetic.h"
+#include "../../shared/precision_type.h"
 #include "function_pool.h"
 #include "fuse_shim.h"
 #include "hip/hip_runtime_api.h"
@@ -301,6 +302,10 @@ bool NodeFactory::CheckLarge1DMaps()
 // Checks whether the non-pow2 length input is supported for a Bluestein compute scheme
 bool NodeFactory::NonPow2LengthSupported(rocfft_precision precision, size_t length)
 {
+    // assume half precision behaves the same as single
+    if(precision == rocfft_precision_half)
+        precision = rocfft_precision_single;
+
     // Exceptions which have been found to perform poorly when compared to the next pow2 length
     static const std::map<rocfft_precision, std::set<size_t>> length_exceptions
         = {{rocfft_precision_single,
@@ -309,7 +314,7 @@ bool NodeFactory::NonPow2LengthSupported(rocfft_precision precision, size_t leng
             {104,  108,  180,  224,  225,  432,  450,  810,  2401,  2430,  2700,  2880,   3125,
              3200, 3240, 3375, 3456, 3600, 3645, 4913, 6561, 11200, 53248, 57344, 106496, 114688}}};
 
-    if(length_exceptions.at(precision).count(length))
+    if(length_excepted(length_exceptions, precision, length))
         return false;
 
     // Look for regular Stockham kernels support
@@ -329,7 +334,12 @@ bool NodeFactory::NonPow2LengthSupported(rocfft_precision precision, size_t leng
     return false;
 }
 
-inline bool SupportedLength(rocfft_precision precision, size_t len)
+size_t NodeFactory::GetBluesteinLength(rocfft_precision precision, size_t len)
+{
+    return BluesteinNode::FindBlue(len, precision, BluesteinSingleNode::SizeFits(len, precision));
+}
+
+bool NodeFactory::SupportedLength(rocfft_precision precision, size_t len)
 {
     // do we have an explicit kernel?
     if(function_pool::has_function(fpkey(len, precision)))
@@ -467,13 +477,41 @@ std::unique_ptr<TreeNode> NodeFactory::CreateNodeFromScheme(ComputeScheme s, Tre
     }
 }
 
-std::unique_ptr<TreeNode> NodeFactory::CreateExplicitNode(NodeMetaData& nodeData, TreeNode* parent)
+std::unique_ptr<TreeNode> NodeFactory::CreateExplicitNode(NodeMetaData& nodeData,
+                                                          TreeNode*     parent,
+                                                          ComputeScheme determined_scheme)
 {
-    // TreeNode*     p = dummyNode->parent;
-    ComputeScheme s = DecideNodeScheme(nodeData, parent);
-    if(s == CS_NONE)
+    // when creating tree from solution map, scheme is L1D but not root
+    // NB:
+    //   Why we need this:
+    //   Ideally, decide-scheme functions don't assign/change any lengths data,
+    //   it should be assigned before. But that is not the case for L1D, when
+    //   deciding L1D scheme, it appends an extra temporary length indicating
+    //   how to "factorize" the large 1D, and then pop in later in "build-tree."
+    //   But when we are creating tree from solution map, we already have the
+    //   determined_scheme; however, if we don't do the decide-node-scheme,
+    //   we lose that factor-length, and the later "pop" casues error. So
+    //   we should still call the decide-node-scheme here as long as we know
+    //   it's a L1D. (But not for the root-node, root-node calls the decide
+    //   function anyway, before we try looking up solutions)
+    if((determined_scheme == CS_L1D_TRTRT || determined_scheme == CS_L1D_CC
+        || determined_scheme == CS_L1D_CRT)
+       && (parent != nullptr))
+    {
+        auto s = DecideNodeScheme(nodeData, parent);
+        if(determined_scheme != s)
+            throw std::runtime_error("solution map error for L1D sub-problem");
+    }
+
+    // createing tree without solution map, must call DecideNodeScheme
+    if(determined_scheme == CS_NONE)
+        determined_scheme = DecideNodeScheme(nodeData, parent);
+
+    // check if successfully created
+    if(determined_scheme == CS_NONE)
         throw std::runtime_error("DecideNodeScheme Failed!: CS_NONE");
-    auto node = CreateNodeFromScheme(s, parent);
+
+    auto node = CreateNodeFromScheme(determined_scheme, parent);
     node->CopyNodeData(nodeData);
     return node;
 }
@@ -571,7 +609,8 @@ ComputeScheme NodeFactory::Decide1DScheme(NodeMetaData& nodeData)
         if(nodeData.length[0] <= block_threshold)
         {
             // Enable block compute under these conditions
-            if(nodeData.precision == rocfft_precision_single)
+            if(nodeData.precision == rocfft_precision_single
+               || nodeData.precision == rocfft_precision_half)
             {
                 if(map1DLengthSingle.find(nodeData.length[0]) != map1DLengthSingle.end())
                 {
@@ -635,7 +674,8 @@ ComputeScheme NodeFactory::Decide1DScheme(NodeMetaData& nodeData)
     }
     else // if not Pow2
     {
-        if(nodeData.precision == rocfft_precision_single)
+        if(nodeData.precision == rocfft_precision_single
+           || nodeData.precision == rocfft_precision_half)
         {
             if(map1DLengthSingle.find(nodeData.length[0]) != map1DLengthSingle.end())
             {
@@ -779,7 +819,8 @@ ComputeScheme NodeFactory::Decide3DScheme(NodeMetaData& nodeData)
         std::map<rocfft_precision, std::set<size_t>> exceptions
             = {{rocfft_precision_single, {84, 112, 168}},
                {rocfft_precision_double, {84, 108, 112, 168}}};
-        if(childScheme == CS_2D_RC && exceptions.at(nodeData.precision).count(nodeData.length[1])
+        if(childScheme == CS_2D_RC
+           && length_excepted(exceptions, nodeData.precision, nodeData.length[1])
            && nodeData.rootIsC2C)
         {
             return CS_3D_TRTRTR;
@@ -832,7 +873,7 @@ bool NodeFactory::use_CS_2D_SINGLE(NodeMetaData& nodeData)
         fpkey(nodeData.length[0], nodeData.length[1], nodeData.precision, CS_KERNEL_2D_SINGLE));
 
     int ldsUsage = nodeData.length[0] * nodeData.length[1] * kernel.transforms_per_block
-                   * sizeof_precision(nodeData.precision);
+                   * complex_type_size(nodeData.precision);
     if(1.5 * ldsUsage > ldsSize)
         return false;
 
@@ -859,8 +900,8 @@ size_t NodeFactory::count_3D_SBRC_nodes(NodeMetaData& nodeData)
         if(function_pool::has_SBRC_kernel(nodeData.length[i], nodeData.precision))
         {
             // make sure the SBRC kernel on that dimension would be tile-aligned
-            auto kernel = function_pool::get_kernel(
-                fpkey(nodeData.length[i], nodeData.precision, CS_KERNEL_STOCKHAM_BLOCK_RC));
+            auto kernel = function_pool::get_kernel(fpkey(
+                nodeData.length[i], nodeData.precision, CS_KERNEL_STOCKHAM_BLOCK_RC, TILE_ALIGNED));
             if(nodeData.length[(i + 2) % nodeData.length.size()] % kernel.transforms_per_block == 0)
                 ++sbrc_dimensions;
         }
