@@ -1,4 +1,4 @@
-// Copyright (C) 2016 - 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2016 - 2023 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -33,6 +33,7 @@
 #include "plan.h"
 #include "rtc_kernel.h"
 #include "transform.h"
+#include "tuning_helper.h"
 
 #include "kernel_launch.h"
 
@@ -42,9 +43,11 @@
 #include "real2complex.h"
 
 #include "../../shared/environment.h"
+#include "../../shared/precision_type.h"
 #include "../../shared/printbuffer.h"
 #include "../../shared/ptrdiff.h"
-#include "rocfft_hip.h"
+#include "../../shared/rocfft_complex.h"
+#include "../../shared/rocfft_hip.h"
 
 // This function is called during creation of plan: enqueue the HIP kernels by function
 // pointers. Return true if everything goes well. Any internal device memory allocation
@@ -53,7 +56,7 @@ bool PlanPowX(ExecPlan& execPlan)
 {
     for(const auto& node : execPlan.execSeq)
     {
-        if(node->CreateTwiddleTableResource() == false)
+        if(node->CreateDeviceResources() == false)
             return false;
 
         if(node->CreateDevKernelArgs() == false)
@@ -88,15 +91,82 @@ bool PlanPowX(ExecPlan& execPlan)
     return true;
 }
 
+bool GetTuningKernelInfo(ExecPlan& execPlan)
+{
+    auto tuningPacket = TuningBenchmarker::GetSingleton().GetPacket();
+    if(!tuningPacket)
+        return false;
+
+    for(size_t i = 0; i < execPlan.execSeq.size(); ++i)
+    {
+        RTCKernel*   localCompiledKernel = execPlan.execSeq[i]->compiledKernel.get().get();
+        GridParam    gp                  = execPlan.gridParam[i];
+        FMKey        key                 = execPlan.execSeq[i]->GetKernelKey();
+        auto         lengths             = std::get<0>(key);
+        KernelConfig config              = std::get<4>(key);
+
+        // get occupancy: 0 means it's compiled (AOT)
+        //               -1 means failed on getting occupancy
+        // TODO- get occupancy of non-RTCKernel
+        int occupancy = 0;
+        if(localCompiledKernel)
+        {
+            // if queried occupancy = 0, which is very likely that this kernel
+            // can't be loaded
+            if(!localCompiledKernel->get_occupancy(
+                   {gp.wgs_x, gp.wgs_y, gp.wgs_z}, gp.lds_bytes, occupancy)
+               || occupancy == 0)
+                occupancy = -1;
+        }
+
+        // factors as string, we will output this to CSV file,
+        // and carry this from phase-0 to phase-1
+        std::string factors_str = "[";
+        std::string COMMA       = "";
+        for(auto factor : config.factors)
+        {
+            factors_str += COMMA + std::to_string(factor);
+            COMMA = ", ";
+        }
+        factors_str += "]";
+
+        // utilization info as string, we will output this to CSV file
+        std::stringstream util_ss;
+        util_ss << "[";
+        util_ss.precision(4);
+        float util_rate = 0.0f;
+        for(auto width : config.factors)
+        {
+            float height = static_cast<float>(lengths[0]) / width / config.threads_per_transform[0];
+            util_ss << std::fixed << height << ", ";
+
+            util_rate += height;
+        }
+        util_rate /= config.factors.size();
+        util_ss << std::fixed << util_rate << "]";
+
+        tuningPacket->num_of_blocks[i] = gp.b_x;
+        tuningPacket->lds_bytes[i]     = gp.lds_bytes;
+        tuningPacket->occupancy[i]     = occupancy;
+        tuningPacket->wgs[i]           = config.workgroup_size;
+        tuningPacket->tpt[i]           = config.threads_per_transform[0];
+        tuningPacket->tpb[i]           = config.transforms_per_block;
+        tuningPacket->util_rate[i]     = util_ss.str();
+        tuningPacket->factors_str[i]   = factors_str;
+    }
+
+    return true;
+}
+
 static size_t data_size_bytes(const std::vector<size_t>& lengths,
                               rocfft_precision           precision,
                               rocfft_array_type          type)
 {
     // first compute the raw number of elements
-    size_t elems = std::accumulate(
+    const size_t elems = std::accumulate(
         lengths.begin(), lengths.end(), static_cast<size_t>(1), std::multiplies<size_t>());
     // size of each element
-    size_t elemsize = (precision == rocfft_precision_single ? sizeof(float) : sizeof(double));
+    const size_t elemsize = real_type_size(precision);
     switch(type)
     {
     case rocfft_array_type_complex_interleaved:
@@ -173,7 +243,7 @@ void DebugPrintBuffer(rocfft_ostream&            stream,
 {
     const size_t size_elems = compute_ptrdiff(length_cm, stride_cm, batch, dist);
 
-    size_t base_type_size = (precision == rocfft_precision_double) ? sizeof(double) : sizeof(float);
+    size_t base_type_size = real_type_size(precision);
     if(type != rocfft_array_type_real)
     {
         // complex elements
@@ -186,14 +256,14 @@ void DebugPrintBuffer(rocfft_ostream&            stream,
     auto stride_rm = stride_cm;
     std::reverse(length_rm.begin(), length_rm.end());
     std::reverse(stride_rm.begin(), stride_rm.end());
-    std::vector<std::vector<char>> bufvec;
-    std::vector<size_t>            print_offset(2, offset);
+    std::vector<hostbuf> bufvec;
+    std::vector<size_t>  print_offset(2, offset);
     if(array_type_is_planar(type))
     {
         // separate the real/imag data, so printbuffer will print them separately
         bufvec.resize(2);
-        bufvec.front().resize(size_bytes / 2);
-        bufvec.back().resize(size_bytes / 2);
+        bufvec.front().alloc(size_bytes / 2);
+        bufvec.back().alloc(size_bytes / 2);
         if(hipMemcpy(bufvec.front().data(), buffer[0], size_bytes / 2, hipMemcpyDeviceToHost)
            != hipSuccess)
             throw std::runtime_error("hipMemcpy failure");
@@ -203,6 +273,12 @@ void DebugPrintBuffer(rocfft_ostream&            stream,
 
         switch(precision)
         {
+        case rocfft_precision_half:
+        {
+            buffer_printer<_Float16> s;
+            s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+            break;
+        }
         case rocfft_precision_single:
         {
             buffer_printer<float> s;
@@ -220,13 +296,35 @@ void DebugPrintBuffer(rocfft_ostream&            stream,
     else
     {
         bufvec.resize(1);
-        bufvec.front().resize(size_bytes);
+        bufvec.front().alloc(size_bytes);
         if(hipMemcpy(bufvec.front().data(), buffer[0], size_bytes, hipMemcpyDeviceToHost)
            != hipSuccess)
             throw std::runtime_error("hipMemcpy failure");
 
         switch(precision)
         {
+        case rocfft_precision_half:
+        {
+            switch(type)
+            {
+            case rocfft_array_type_complex_interleaved:
+            case rocfft_array_type_hermitian_interleaved:
+            {
+                buffer_printer<rocfft_complex<_Float16>> s;
+                s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+                break;
+            }
+            case rocfft_array_type_real:
+            {
+                buffer_printer<_Float16> s;
+                s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+                break;
+            }
+            default:
+                throw std::runtime_error("invalid array format");
+            }
+            break;
+        }
         case rocfft_precision_single:
         {
             switch(type)
@@ -234,7 +332,7 @@ void DebugPrintBuffer(rocfft_ostream&            stream,
             case rocfft_array_type_complex_interleaved:
             case rocfft_array_type_hermitian_interleaved:
             {
-                buffer_printer<std::complex<float>> s;
+                buffer_printer<rocfft_complex<float>> s;
                 s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
                 break;
             }
@@ -256,7 +354,7 @@ void DebugPrintBuffer(rocfft_ostream&            stream,
             case rocfft_array_type_complex_interleaved:
             case rocfft_array_type_hermitian_interleaved:
             {
-                buffer_printer<std::complex<double>> s;
+                buffer_printer<rocfft_complex<double>> s;
                 s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
                 break;
             }
@@ -295,27 +393,69 @@ void SetDefaultCallback(const TreeNode* node, const SetCallbackType& type, void*
 
     if(is_complex && type == SetCallbackType::LOAD)
     {
-        result = (node->precision == rocfft_precision_single)
-                     ? hipMemcpyFromSymbol(cb, HIP_SYMBOL(load_cb_default_float2), sizeof(void*))
-                     : hipMemcpyFromSymbol(cb, HIP_SYMBOL(load_cb_default_double2), sizeof(void*));
+        switch(node->precision)
+        {
+        case rocfft_precision_half:
+            result
+                = hipMemcpyFromSymbol(cb, HIP_SYMBOL(load_cb_default_complex_half), sizeof(void*));
+            break;
+        case rocfft_precision_single:
+            result
+                = hipMemcpyFromSymbol(cb, HIP_SYMBOL(load_cb_default_complex_float), sizeof(void*));
+            break;
+        case rocfft_precision_double:
+            result = hipMemcpyFromSymbol(
+                cb, HIP_SYMBOL(load_cb_default_complex_double), sizeof(void*));
+            break;
+        }
     }
     else if(is_complex && type == SetCallbackType::STORE)
     {
-        result = (node->precision == rocfft_precision_single)
-                     ? hipMemcpyFromSymbol(cb, HIP_SYMBOL(store_cb_default_float2), sizeof(void*))
-                     : hipMemcpyFromSymbol(cb, HIP_SYMBOL(store_cb_default_double2), sizeof(void*));
+        switch(node->precision)
+        {
+        case rocfft_precision_half:
+            result
+                = hipMemcpyFromSymbol(cb, HIP_SYMBOL(store_cb_default_complex_half), sizeof(void*));
+            break;
+        case rocfft_precision_single:
+            result = hipMemcpyFromSymbol(
+                cb, HIP_SYMBOL(store_cb_default_complex_float), sizeof(void*));
+            break;
+        case rocfft_precision_double:
+            result = hipMemcpyFromSymbol(
+                cb, HIP_SYMBOL(store_cb_default_complex_double), sizeof(void*));
+            break;
+        }
     }
     else if(!is_complex && type == SetCallbackType::LOAD)
     {
-        result = (node->precision == rocfft_precision_single)
-                     ? hipMemcpyFromSymbol(cb, HIP_SYMBOL(load_cb_default_float), sizeof(void*))
-                     : hipMemcpyFromSymbol(cb, HIP_SYMBOL(load_cb_default_double), sizeof(void*));
+        switch(node->precision)
+        {
+        case rocfft_precision_half:
+            result = hipMemcpyFromSymbol(cb, HIP_SYMBOL(load_cb_default_half), sizeof(void*));
+            break;
+        case rocfft_precision_single:
+            result = hipMemcpyFromSymbol(cb, HIP_SYMBOL(load_cb_default_float), sizeof(void*));
+            break;
+        case rocfft_precision_double:
+            result = hipMemcpyFromSymbol(cb, HIP_SYMBOL(load_cb_default_double), sizeof(void*));
+            break;
+        }
     }
     else if(!is_complex && type == SetCallbackType::STORE)
     {
-        result = (node->precision == rocfft_precision_single)
-                     ? hipMemcpyFromSymbol(cb, HIP_SYMBOL(store_cb_default_float), sizeof(void*))
-                     : hipMemcpyFromSymbol(cb, HIP_SYMBOL(store_cb_default_double), sizeof(void*));
+        switch(node->precision)
+        {
+        case rocfft_precision_half:
+            result = hipMemcpyFromSymbol(cb, HIP_SYMBOL(store_cb_default_half), sizeof(void*));
+            break;
+        case rocfft_precision_single:
+            result = hipMemcpyFromSymbol(cb, HIP_SYMBOL(store_cb_default_float), sizeof(void*));
+            break;
+        case rocfft_precision_double:
+            result = hipMemcpyFromSymbol(cb, HIP_SYMBOL(store_cb_default_double), sizeof(void*));
+            break;
+        }
     }
 
     if(result != hipSuccess)
@@ -332,12 +472,15 @@ void TransformPowX(const ExecPlan&       execPlan,
     assert(execPlan.execSeq.size() == execPlan.devFnCall.size());
     assert(execPlan.execSeq.size() == execPlan.gridParam.size());
 
+    bool processing_tuning = TuningBenchmarker::GetSingleton().IsProcessingTuning();
+    auto tuningPacket      = TuningBenchmarker::GetSingleton().GetPacket();
     // we can log profile information if we're on the null stream,
     // since we will be able to wait for the transform to finish
-    bool            emit_profile_log  = LOG_PROFILE_ENABLED() && !info->rocfft_stream;
-    bool            emit_kernelio_log = LOG_KERNELIO_ENABLED();
-    rocfft_ostream* kernelio_stream   = nullptr;
-    float           max_memory_bw     = 0.0;
+    bool emit_profile_log  = (processing_tuning || LOG_PROFILE_ENABLED()) && !info->rocfft_stream;
+    bool emit_kernelio_log = LOG_KERNELIO_ENABLED();
+
+    rocfft_ostream* kernelio_stream = nullptr;
+    float           max_memory_bw   = 0.0;
     hipEvent_t      start, stop;
     if(emit_profile_log)
     {
@@ -372,9 +515,7 @@ void TransformPowX(const ExecPlan&       execPlan,
             data.log_func = nullptr;
 
         // Size of complex type
-        const size_t complexTSize = (data.node->precision == rocfft_precision_single)
-                                        ? sizeof(float) * 2
-                                        : sizeof(double) * 2;
+        const size_t complexTSize = complex_type_size(data.node->precision);
 
         switch(data.node->obIn)
         {
@@ -618,6 +759,9 @@ void TransformPowX(const ExecPlan&       execPlan,
                 auto efficiency_pct = 0.0;
                 if(max_memory_bw != 0.0)
                     efficiency_pct = 100.0 * exec_bw / max_memory_bw;
+                if(processing_tuning)
+                    tuningPacket->bw_effs[i] = efficiency_pct;
+
                 log_profile(__func__,
                             "scheme",
                             PrintScheme(execPlan.execSeq[i]->scheme),
