@@ -31,6 +31,7 @@ using namespace std::placeholders;
 #include "rtc_cache.h"
 #include "rtc_realcomplex_gen.h"
 #include "rtc_stockham_gen.h"
+#include "rtc_twiddle_gen.h"
 
 #include "device/kernel-generator-embed.h"
 
@@ -133,18 +134,15 @@ void stockham_combo(ComputeScheme             scheme,
         unitstride_range = {false};
 
         base_steps.resize(1);
-        // All SBRCs have TILE_UNALIGNED
-        sbrc_trans_types.push_back(SBRC_TRANSPOSE_TYPE::TILE_UNALIGNED);
-        // Finish SBRC-2D
-        if(scheme == CS_KERNEL_STOCKHAM_BLOCK_RC)
-            break;
-        // All 3D SBRCs have TILE_ALIGNED, but "NO" SBRC_TRANSPOSE_TYPE::NONE
-        sbrc_trans_types.push_back(SBRC_TRANSPOSE_TYPE::TILE_ALIGNED);
+        // All SBRCs have ALIGNED and UNALIGNED, but no NONE
         sbrc_trans_types.erase(sbrc_trans_types.begin());
-        // Finish ERC
-        if(scheme == CS_KERNEL_STOCKHAM_R_TO_CMPLX_TRANSPOSE_Z_XY)
+        sbrc_trans_types.push_back(SBRC_TRANSPOSE_TYPE::TILE_ALIGNED);
+        sbrc_trans_types.push_back(SBRC_TRANSPOSE_TYPE::TILE_UNALIGNED);
+        // Finish SBRC-2D and SBRC-3D-ERC without DIAGONAL
+        if(scheme == CS_KERNEL_STOCKHAM_BLOCK_RC
+           || scheme == CS_KERNEL_STOCKHAM_R_TO_CMPLX_TRANSPOSE_Z_XY)
             break;
-        // DIAGNAL Transpose
+        // DIAGONAL Transpose
         sbrc_trans_types.push_back(SBRC_TRANSPOSE_TYPE::DIAGONAL);
 
         break;
@@ -227,6 +225,9 @@ void build_stockham_function_pool(CompileQueue& queue)
     // scaling Stockham kernels are always built at runtime
     const bool enable_scaling = false;
 
+    // fused Bluestein kernels are also always built at runtime
+    auto fuseBlue = BluesteinFuseType::BFT_NONE;
+
     for(const auto& i : fp.get_map())
     {
         // we only want to compile kernels explicitly marked for AOT RTC
@@ -240,6 +241,15 @@ void build_stockham_function_pool(CompileQueue& queue)
         std::vector<unsigned int> factors;
         std::copy(i.second.factors.begin(), i.second.factors.end(), std::back_inserter(factors));
 
+        StockhamGeneratorSpecs specs{factors,
+                                     {},
+                                     {static_cast<unsigned int>(precision)},
+                                     static_cast<unsigned int>(i.second.workgroup_size),
+                                     PrintScheme(scheme)};
+        specs.threads_per_transform = i.second.threads_per_transform[0];
+        specs.half_lds              = i.second.half_lds;
+        specs.direct_to_from_reg    = i.second.direct_to_from_reg;
+
         stockham_combo(scheme,
                        i.second,
                        [=, &queue](int                     direction,
@@ -274,10 +284,9 @@ void build_stockham_function_pool(CompileQueue& queue)
                                    return;
                            }
 
-                           auto kernel_name = stockham_rtc_kernel_name(scheme,
-                                                                       length1D,
-                                                                       0,
-                                                                       0,
+                           auto kernel_name = stockham_rtc_kernel_name(specs,
+                                                                       specs,
+                                                                       scheme,
                                                                        direction,
                                                                        precision,
                                                                        placement,
@@ -292,7 +301,8 @@ void build_stockham_function_pool(CompileQueue& queue)
                                                                        intrinsic,
                                                                        sbrc_trans_type,
                                                                        callbacks,
-                                                                       enable_scaling);
+                                                                       enable_scaling,
+                                                                       fuseBlue);
                            std::function<std::string(const std::string&)> generate_src
                                = [=](const std::string& kernel_name) -> std::string {
                                StockhamGeneratorSpecs specs{
@@ -323,7 +333,8 @@ void build_stockham_function_pool(CompileQueue& queue)
                                                    intrinsic,
                                                    sbrc_trans_type,
                                                    callbacks,
-                                                   enable_scaling);
+                                                   enable_scaling,
+                                                   fuseBlue);
                            };
                            queue.push({kernel_name, generate_src});
                        });
@@ -415,6 +426,28 @@ void build_apply_callback(CompileQueue& queue)
     }
 }
 
+void build_twiddle(CompileQueue& queue)
+{
+    const auto twiddle_kernel_types = {
+        TwiddleTableType::RADICES,
+        TwiddleTableType::LENGTH_N,
+        TwiddleTableType::HALF_N,
+        TwiddleTableType::LARGE,
+    };
+    for(auto precision : {rocfft_precision_single, rocfft_precision_double})
+    {
+        for(auto type : twiddle_kernel_types)
+        {
+            auto kernel_name = twiddle_rtc_kernel_name(type, precision);
+            std::function<std::string(const std::string&)> generate_src
+                = [=](const std::string& kernel_name) -> std::string {
+                return twiddle_rtc(kernel_name, type, precision);
+            };
+            queue.push({kernel_name, generate_src});
+        }
+    }
+}
+
 int main(int argc, char** argv)
 {
     if(argc < 5)
@@ -474,6 +507,7 @@ int main(int argc, char** argv)
     build_stockham_function_pool(queue);
     build_realcomplex(queue);
     build_apply_callback(queue);
+    build_twiddle(queue);
 
     // signal end of results with empty work items
     for(size_t i = 0; i < NUM_THREADS; ++i)
