@@ -1,4 +1,4 @@
-// Copyright (C) 2021 - 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2021 - 2023 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -19,17 +19,21 @@
 // THE SOFTWARE.
 
 #include "tree_node_1D.h"
+#include "../../shared/precision_type.h"
 #include "../device/kernels/bank_shift.h"
 #include "function_pool.h"
 #include "fuse_shim.h"
 #include "node_factory.h"
+#include "tuning_helper.h"
 #include <numeric>
 
 /*****************************************************
  * L1D_TRTRT  *
  *****************************************************/
-void TRTRT1DNode::BuildTree_internal()
+void TRTRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
 {
+    bool noSolution = child_schemes.empty();
+
     size_t lenFactor1 = length.back();
     size_t lenFactor0 = length[0] / lenFactor1;
     if(lenFactor0 * lenFactor1 != length[0])
@@ -37,6 +41,8 @@ void TRTRT1DNode::BuildTree_internal()
     length.pop_back();
 
     // first transpose
+    if(!noSolution)
+        assert(child_schemes[0] == CS_KERNEL_TRANSPOSE);
     auto trans1Plan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE, this);
     trans1Plan->length.push_back(lenFactor0);
     trans1Plan->length.push_back(lenFactor1);
@@ -56,11 +62,16 @@ void TRTRT1DNode::BuildTree_internal()
     {
         row1PlanData.length.push_back(length[index]);
     }
-    auto row1Plan     = NodeFactory::CreateExplicitNode(row1PlanData, this);
-    row1Plan->large1D = 0;
+
+    // skip the decide scheme part in node factory
+    ComputeScheme determined_scheme = (noSolution) ? CS_NONE : child_schemes[1];
+    auto          row1Plan = NodeFactory::CreateExplicitNode(row1PlanData, this, determined_scheme);
+    row1Plan->large1D      = 0;
     row1Plan->RecursiveBuildTree();
 
     // second transpose
+    if(!noSolution)
+        assert(child_schemes[2] == CS_KERNEL_TRANSPOSE);
     auto trans2Plan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE, this);
     trans2Plan->length.push_back(lenFactor1);
     trans2Plan->length.push_back(lenFactor0);
@@ -73,6 +84,8 @@ void TRTRT1DNode::BuildTree_internal()
     trans2Plan->SetTransposeOutputLength();
 
     // second row fft
+    if(!noSolution)
+        assert(child_schemes[3] == CS_KERNEL_STOCKHAM);
     auto row2Plan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM, this);
     row2Plan->length.push_back(lenFactor0);
     row2Plan->length.push_back(lenFactor1);
@@ -83,6 +96,8 @@ void TRTRT1DNode::BuildTree_internal()
     }
 
     // third transpose
+    if(!noSolution)
+        assert(child_schemes[4] == CS_KERNEL_TRANSPOSE);
     auto trans3Plan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE, this);
     trans3Plan->length.push_back(lenFactor0);
     trans3Plan->length.push_back(lenFactor1);
@@ -296,8 +311,10 @@ void TRTRT1DNode::AssignParams_internal()
 /*****************************************************
  * L1D_CC  *
  *****************************************************/
-void CC1DNode::BuildTree_internal()
+void CC1DNode::BuildTree_internal(const SchemeVec& child_schemes)
 {
+    bool noSolution = child_schemes.empty();
+
     size_t lenFactor1 = length.back();
     size_t lenFactor0 = length[0] / lenFactor1;
     if(lenFactor0 * lenFactor1 != length[0])
@@ -305,18 +322,22 @@ void CC1DNode::BuildTree_internal()
     length.pop_back();
 
     // first plan, column-to-column
+    if(!noSolution)
+        assert(child_schemes[0] == CS_KERNEL_STOCKHAM_BLOCK_CC);
     auto col2colPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CC, this);
+
+    col2colPlan->typeBlue = typeBlue;
+    col2colPlan->fuseBlue = fuseBlue;
+    if(fuseBlue != BFT_NONE)
+    {
+        col2colPlan->lengthBlue  = lengthBlue;
+        col2colPlan->lengthBlueN = lengthBlueN;
+        if(fuseBlue == BFT_FWD_CHIRP || fuseBlue == BFT_FWD_CHIRP_MUL)
+            col2colPlan->need_chirp = true;
+    }
+
     // large1D flag to confirm we need multiply twiddle factor
     col2colPlan->large1D = length[0];
-    if(function_pool::has_SBCC_kernel(lenFactor1, precision))
-    {
-        // decompose the large twd table for L1D_CC
-        // exclude some exceptions that don't get benefit from 3-step LargeTwd (set in FFTKernel)
-        auto kernel
-            = function_pool::get_kernel(fpkey(lenFactor1, precision, CS_KERNEL_STOCKHAM_BLOCK_CC));
-        col2colPlan->largeTwd3Steps = kernel.use_3steps_large_twd;
-        col2colPlan->set_large_twd_base_steps(length[0]);
-    }
     col2colPlan->length.push_back(lenFactor1);
     col2colPlan->length.push_back(lenFactor0);
     col2colPlan->dimension = 1;
@@ -328,7 +349,20 @@ void CC1DNode::BuildTree_internal()
     std::swap(col2colPlan->outputLength[0], col2colPlan->outputLength[1]);
 
     // second plan, row-to-column
+    if(!noSolution)
+        assert(child_schemes[1] == CS_KERNEL_STOCKHAM_BLOCK_RC);
     auto row2colPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_RC, this);
+
+    row2colPlan->typeBlue = typeBlue;
+    row2colPlan->fuseBlue = fuseBlue;
+    if(fuseBlue != BFT_NONE)
+    {
+        row2colPlan->lengthBlue  = lengthBlue;
+        row2colPlan->lengthBlueN = lengthBlueN;
+        if(fuseBlue == BFT_INV_CHIRP_MUL)
+            row2colPlan->need_chirp = true;
+    }
+
     row2colPlan->length.push_back(lenFactor0);
     row2colPlan->length.push_back(lenFactor1);
     row2colPlan->dimension = 1;
@@ -349,7 +383,10 @@ void CC1DNode::AssignParams_internal()
     auto& col2colPlan = childNodes[0];
     auto& row2colPlan = childNodes[1];
 
-    if((obOut == OB_USER_OUT) || (obOut == OB_TEMP_CMPLX_FOR_REAL))
+    assert(inStrideBlue.size() == outStrideBlue.size());
+    bool setBlueData = inStrideBlue.size();
+
+    if((obOut == OB_USER_OUT) || (obOut == OB_TEMP_CMPLX_FOR_REAL) || (obOut == OB_TEMP_BLUESTEIN))
     {
         // B -> T
         col2colPlan->inStride.push_back(inStride[0] * col2colPlan->length[1]);
@@ -360,11 +397,29 @@ void CC1DNode::AssignParams_internal()
         col2colPlan->outStride.push_back(1);
         col2colPlan->oDist = length[0];
 
+        if(setBlueData)
+        {
+            col2colPlan->outStrideBlue.push_back(col2colPlan->length[1]);
+            col2colPlan->outStrideBlue.push_back(1);
+            col2colPlan->oDistBlue = lengthBlue;
+
+            col2colPlan->inStrideBlue.push_back(inStrideBlue[0] * col2colPlan->length[1]);
+            col2colPlan->inStrideBlue.push_back(inStrideBlue[0]);
+            col2colPlan->iDistBlue = iDistBlue;
+        }
+
         for(size_t index = 1; index < length.size(); index++)
         {
             col2colPlan->inStride.push_back(inStride[index]);
             col2colPlan->outStride.push_back(col2colPlan->oDist);
             col2colPlan->oDist *= length[index];
+
+            if(setBlueData)
+            {
+                col2colPlan->inStrideBlue.push_back(inStrideBlue[index]);
+                col2colPlan->outStrideBlue.push_back(col2colPlan->oDistBlue);
+                col2colPlan->oDistBlue *= length[index];
+            }
         }
 
         // T -> B
@@ -376,11 +431,29 @@ void CC1DNode::AssignParams_internal()
         row2colPlan->outStride.push_back(outStride[0] * row2colPlan->length[1]);
         row2colPlan->oDist = oDist;
 
+        if(setBlueData)
+        {
+            row2colPlan->inStrideBlue.push_back(1);
+            row2colPlan->inStrideBlue.push_back(row2colPlan->length[0]);
+            row2colPlan->iDistBlue = lengthBlue;
+
+            row2colPlan->outStrideBlue.push_back(outStrideBlue[0]);
+            row2colPlan->outStrideBlue.push_back(outStrideBlue[0] * row2colPlan->length[1]);
+            row2colPlan->oDistBlue = oDistBlue;
+        }
+
         for(size_t index = 1; index < length.size(); index++)
         {
             row2colPlan->inStride.push_back(row2colPlan->iDist);
             row2colPlan->iDist *= length[index];
             row2colPlan->outStride.push_back(outStride[index]);
+
+            if(setBlueData)
+            {
+                row2colPlan->inStrideBlue.push_back(row2colPlan->iDistBlue);
+                row2colPlan->iDistBlue *= length[index];
+                row2colPlan->outStrideBlue.push_back(outStrideBlue[index]);
+            }
         }
     }
     else
@@ -396,9 +469,21 @@ void CC1DNode::AssignParams_internal()
         col2colPlan->inStride.push_back(inStride[0]);
         col2colPlan->iDist = iDist;
 
+        if(setBlueData)
+        {
+            col2colPlan->inStrideBlue.push_back(inStrideBlue[0] * col2colPlan->length[1]);
+            col2colPlan->inStrideBlue.push_back(inStrideBlue[0]);
+            col2colPlan->iDistBlue = iDistBlue;
+        }
+
         for(size_t index = 1; index < length.size(); index++)
+        {
             col2colPlan->inStride.push_back(inStride[index]);
 
+            if(setBlueData)
+                col2colPlan->inStrideBlue.push_back(inStrideBlue[index]);
+        }
+
         if(parent->scheme == CS_L1D_TRTRT)
         {
             col2colPlan->outStride.push_back(parent->outStride[0] * col2colPlan->length[1]);
@@ -407,23 +492,58 @@ void CC1DNode::AssignParams_internal()
                                              * col2colPlan->length[0]);
             col2colPlan->oDist = parent->oDist;
 
+            if(setBlueData)
+            {
+                col2colPlan->outStrideBlue.push_back(parent->outStrideBlue[0]
+                                                     * col2colPlan->length[1]);
+                col2colPlan->outStrideBlue.push_back(parent->outStrideBlue[0]);
+                col2colPlan->outStrideBlue.push_back(
+                    parent->outStrideBlue[0] * col2colPlan->length[1] * col2colPlan->length[0]);
+                col2colPlan->oDistBlue = parent->oDistBlue;
+            }
+
             for(size_t index = 1; index < parent->length.size(); index++)
+            {
                 col2colPlan->outStride.push_back(parent->outStride[index]);
+
+                if(setBlueData)
+                    col2colPlan->outStrideBlue.push_back(parent->outStrideBlue[index]);
+            }
         }
         else
         {
             // we dont have B info here, need to assume packed data and descended
             // from 2D/3D
-            assert(parent->outStride[0] == 1);
+            //assert(parent->outStride[0] == 1);
+            //assert(parent->outStrideBlue[0] == 1);
 
             col2colPlan->outStride.push_back(col2colPlan->length[1]);
             col2colPlan->outStride.push_back(1);
-            col2colPlan->oDist = col2colPlan->length[1] * col2colPlan->length[0];
+
+            if(setBlueData)
+            {
+                col2colPlan->outStrideBlue.push_back(col2colPlan->length[1]);
+                col2colPlan->outStrideBlue.push_back(1);
+            }
+
+            if(fuseBlue != BFT_NONE)
+            {
+                col2colPlan->oDist     = lengthBlueN;
+                col2colPlan->oDistBlue = lengthBlue;
+            }
+            else
+                col2colPlan->oDist = col2colPlan->length[1] * col2colPlan->length[0];
 
             for(size_t index = 1; index < length.size(); index++)
             {
                 col2colPlan->outStride.push_back(col2colPlan->oDist);
                 col2colPlan->oDist *= length[index];
+
+                if(setBlueData)
+                {
+                    col2colPlan->outStrideBlue.push_back(col2colPlan->oDistBlue);
+                    col2colPlan->oDistBlue *= length[index];
+                }
             }
         }
 
@@ -436,8 +556,23 @@ void CC1DNode::AssignParams_internal()
                                             * row2colPlan->length[1]);
             row2colPlan->iDist = parent->oDist;
 
+            if(setBlueData)
+            {
+                row2colPlan->inStrideBlue.push_back(parent->outStrideBlue[0]);
+                row2colPlan->inStrideBlue.push_back(parent->outStrideBlue[0]
+                                                    * row2colPlan->length[0]);
+                row2colPlan->inStrideBlue.push_back(
+                    parent->outStrideBlue[0] * row2colPlan->length[0] * row2colPlan->length[1]);
+                row2colPlan->iDistBlue = parent->oDistBlue;
+            }
+
             for(size_t index = 1; index < parent->length.size(); index++)
+            {
                 row2colPlan->inStride.push_back(parent->outStride[index]);
+
+                if(setBlueData)
+                    row2colPlan->inStrideBlue.push_back(parent->outStrideBlue[index]);
+            }
         }
         else
         {
@@ -445,12 +580,31 @@ void CC1DNode::AssignParams_internal()
             // from 2D/3D
             row2colPlan->inStride.push_back(1);
             row2colPlan->inStride.push_back(row2colPlan->length[0]);
-            row2colPlan->iDist = row2colPlan->length[0] * row2colPlan->length[1];
+
+            if(setBlueData)
+            {
+                row2colPlan->inStrideBlue.push_back(1);
+                row2colPlan->inStrideBlue.push_back(row2colPlan->length[0]);
+            }
+
+            if(fuseBlue != BFT_NONE)
+            {
+                row2colPlan->iDist     = lengthBlueN;
+                row2colPlan->iDistBlue = lengthBlue;
+            }
+            else
+                row2colPlan->iDist = row2colPlan->length[0] * row2colPlan->length[1];
 
             for(size_t index = 1; index < length.size(); index++)
             {
                 row2colPlan->inStride.push_back(row2colPlan->iDist);
                 row2colPlan->iDist *= length[index];
+
+                if(setBlueData)
+                {
+                    row2colPlan->inStrideBlue.push_back(row2colPlan->iDistBlue);
+                    row2colPlan->iDistBlue *= length[index];
+                }
             }
         }
 
@@ -458,8 +612,20 @@ void CC1DNode::AssignParams_internal()
         row2colPlan->outStride.push_back(outStride[0] * row2colPlan->length[1]);
         row2colPlan->oDist = oDist;
 
+        if(setBlueData)
+        {
+            row2colPlan->outStrideBlue.push_back(outStrideBlue[0]);
+            row2colPlan->outStrideBlue.push_back(outStrideBlue[0] * row2colPlan->length[1]);
+            row2colPlan->oDistBlue = oDistBlue;
+        }
+
         for(size_t index = 1; index < length.size(); index++)
+        {
             row2colPlan->outStride.push_back(outStride[index]);
+
+            if(setBlueData)
+                row2colPlan->outStrideBlue.push_back(outStrideBlue[index]);
+        }
     }
 
     // special case for strided large 1D FFT with dist 1
@@ -497,8 +663,10 @@ void CC1DNode::AssignParams_internal()
 /*****************************************************
  * L1D_CRT  *
  *****************************************************/
-void CRT1DNode::BuildTree_internal()
+void CRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
 {
+    bool noSolution = child_schemes.empty();
+
     size_t lenFactor1 = length.back();
     size_t lenFactor0 = length[0] / lenFactor1;
     if(lenFactor0 * lenFactor1 != length[0])
@@ -506,18 +674,11 @@ void CRT1DNode::BuildTree_internal()
     length.pop_back();
 
     // first plan, column-to-column
+    if(!noSolution)
+        assert(child_schemes[0] == CS_KERNEL_STOCKHAM_BLOCK_CC);
     auto col2colPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CC, this);
     // large1D flag to confirm we need multiply twiddle factor
     col2colPlan->large1D = length[0];
-    if(function_pool::has_SBCC_kernel(lenFactor1, precision))
-    {
-        // decompose the large twd table for L1D_CRT
-        // exclude some exceptions that don't get benefit from 3-step LargeTwd (set in FFTKernel)
-        auto kernel
-            = function_pool::get_kernel(fpkey(lenFactor1, precision, CS_KERNEL_STOCKHAM_BLOCK_CC));
-        col2colPlan->largeTwd3Steps = kernel.use_3steps_large_twd;
-        col2colPlan->set_large_twd_base_steps(length[0]);
-    }
     col2colPlan->length.push_back(lenFactor1);
     col2colPlan->length.push_back(lenFactor0);
     col2colPlan->dimension = 1;
@@ -529,6 +690,8 @@ void CRT1DNode::BuildTree_internal()
     std::swap(col2colPlan->outputLength[0], col2colPlan->outputLength[1]);
 
     // second plan, row-to-row
+    if(!noSolution)
+        assert(child_schemes[1] == CS_KERNEL_STOCKHAM);
     auto row2rowPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM, this);
     row2rowPlan->length.push_back(lenFactor0);
     row2rowPlan->length.push_back(lenFactor1);
@@ -541,6 +704,8 @@ void CRT1DNode::BuildTree_internal()
     // row2rowPlan->allowOutofplace = false;
 
     // third plan, transpose
+    if(!noSolution)
+        assert(child_schemes[2] == CS_KERNEL_TRANSPOSE);
     auto transPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE, this);
     transPlan->length.push_back(lenFactor0);
     transPlan->length.push_back(lenFactor1);
@@ -716,15 +881,17 @@ void Stockham1DNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
     for(size_t j = 1; j < length.size(); j++)
         batch_accum *= length[j];
 
-    auto kernel = function_pool::get_kernel(fpkey(length[0], precision));
+    auto key    = GetKernelKey();
+    auto kernel = function_pool::get_kernel(key);
     fnPtr       = kernel.device_function;
 
     if(ebtype != EmbeddedType::NONE)
         lds_padding = 1;
 
     bwd      = kernel.transforms_per_block;
+    wgs      = kernel.workgroup_size;
     gp.b_x   = (batch_accum + bwd - 1) / bwd;
-    gp.wgs_x = kernel.workgroup_size;
+    gp.wgs_x = wgs;
 
     // we don't even need lds (kernel_1,2,3,4,5,6,7,10,11,13,17) since we don't use them at all
     // TODO: we can even use swizzle to do the butterfly shuffle if threads_per_transform[0] <= warpSize
@@ -749,14 +916,18 @@ void Stockham1DNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
     }
 }
 
-bool Stockham1DNode::CreateTwiddleTableResource()
+bool Stockham1DNode::CreateDeviceResources()
 {
     twd_attach_halfN = (ebtype != EmbeddedType::NONE);
-    return LeafNode::CreateTwiddleTableResource();
+    return LeafNode::CreateDeviceResources();
 }
 
 std::vector<size_t> Stockham1DNode::CollapsibleDims()
 {
+    // do not collapse on multi-kernel fused Bluestein nodes
+    if(typeBlue == BT_MULTI_KERNEL_FUSED)
+        return {};
+
     // fastest dim is FFT, the rest is collapsible
     std::vector<size_t> ret(length.size() - 1);
     std::iota(ret.begin(), ret.end(), 1);
@@ -766,21 +937,49 @@ std::vector<size_t> Stockham1DNode::CollapsibleDims()
 /*****************************************************
  * SBCC  *
  *****************************************************/
-bool SBCCNode::KernelCheck()
+bool SBCCNode::KernelCheck(std::vector<FMKey>& kernel_keys)
 {
-    bool res = LeafNode::KernelCheck();
+    bool res = LeafNode::KernelCheck(kernel_keys);
+    if(!res)
+        return false;
 
-    // set according to benchmark
-    SetDirectRegType();
+    if(large1D > 0)
+    {
+        FMKey key      = GetKernelKey();
+        auto  kernel   = function_pool::get_kernel(key);
+        largeTwd3Steps = kernel.use_3steps_large_twd;
+        get_large_twd_base_steps(large1D, largeTwd3Steps, largeTwdBase, ltwdSteps);
+    }
 
-    // set according to benchmark
-    SetIntrinsicMode();
+    // if we are doing tuning or running with the tuned solution, we have the specified_key.
+    // we must directly run the kernel with the exact setting as the config
+    // without the hardcoded tuning
+    if(specified_key != nullptr)
+    {
+        InitIntrinsicMode();
+        return true;
+    }
 
-    return res;
+    // hardocded-tuning according to benchmark
+    TuneDirectRegType();
+
+    // check if we can use buffer instr
+    InitIntrinsicMode();
+    // hardocded-tuning according to benchmark
+    TuneIntrinsicMode();
+
+    return true;
 }
 
-void SBCCNode::SetDirectRegType()
+void SBCCNode::TuneDirectRegType()
 {
+    // half precision has not been tested yet, disable it for now.
+    if(precision == rocfft_precision_half)
+    {
+        dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
+        return;
+    }
+
     // for Navi, Haven't tested all.
     if(is_device_gcn_arch(deviceProp, "gfx1030"))
     {
@@ -796,7 +995,7 @@ void SBCCNode::SetDirectRegType()
         std::map<rocfft_precision, std::set<size_t>> exceptions
             = {{rocfft_precision_single, {125, 192, 216, 224, 240, 243}},
                {rocfft_precision_double, {224, 343}}};
-        if(exceptions.at(precision).count(length[0]))
+        if(length_excepted(exceptions, precision, length[0]))
             dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
     }
     else if(is_device_gcn_arch(deviceProp, "gfx90a"))
@@ -807,69 +1006,99 @@ void SBCCNode::SetDirectRegType()
         std::map<rocfft_precision, std::set<size_t>> exceptions
             = {{rocfft_precision_single, {125, 192, 216, 200, 224, 240}},
                {rocfft_precision_double, {125, 224, 243}}};
-        if(exceptions.at(precision).count(length[0]))
+        if(length_excepted(exceptions, precision, length[0]))
             dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
     }
 }
 
-void SBCCNode::SetIntrinsicMode()
+void SBCCNode::InitIntrinsicMode()
 {
-    // NB: remember set this value at this point instead of SetupGPAndFnPtr_internal()
-    //     since we might need to pass this value to RTC generator
-    intrinsicMode = IntrinsicAccessType::DISABLE_BOTH;
-
-    // TODO- To test on gfx90a
-    if((is_device_gcn_arch(deviceProp, "gfx906") || is_device_gcn_arch(deviceProp, "gfx908")
-        || is_device_gcn_arch(deviceProp, "gfx1030"))
-       && (dir2regMode == TRY_ENABLE_IF_SUPPORT))
+    // 1. General rejections: (Guard) cases we definitely can't use buffer instruction
+    // 2. half precision has not been tested yet, disable it for now.
+    if(((uint64_t)iDist * batch * complex_type_size(precision) >= 0xFFFFFFFF)
+       || ((uint64_t)oDist * batch * complex_type_size(precision) >= 0xFFFFFFFF)
+       || (precision == rocfft_precision_half))
     {
-        // General rejections: cases we can't use buffer load
-        if(((uint64_t)iDist * batch * sizeof_precision(precision) < 0xFFFFFFFF)
-           && ((uint64_t)oDist * batch * sizeof_precision(precision) < 0xFFFFFFFF))
-        {
-            if(placement == rocfft_placement_inplace)
-                intrinsicMode = IntrinsicAccessType::ENABLE_LOAD_ONLY;
-            else
-                intrinsicMode = IntrinsicAccessType::ENABLE_BOTH;
-        }
+        intrinsicMode = IntrinsicAccessType::DISABLE_BOTH;
+        return;
+    }
 
-        // Based on benchmark results
-        if(is_device_gcn_arch(deviceProp, "gfx906"))
-        {
-            // bad results from benchmark:
-            // {96,sp}, {125,sp}, {192,sp/dp}, {240,dp}, {256,sp/dp}, {343,sp/dp}
-            std::map<rocfft_precision, std::set<size_t>> exceptions
-                = {{rocfft_precision_single, {96, 125, 192, 256, 343}},
-                   {rocfft_precision_double, {192, 240, 256, 343}}};
-            if(exceptions.at(precision).count(length[0]))
-                intrinsicMode = IntrinsicAccessType::DISABLE_BOTH;
-        }
-        else if(is_device_gcn_arch(deviceProp, "gfx908"))
-        {
-            // bad results from benchmark:
-            // {104,sp/dp}, {192,dp}, {240,dp}, {289,sp}
-            std::map<rocfft_precision, std::set<size_t>> exceptions = {
-                {rocfft_precision_single, {104, 289}}, {rocfft_precision_double, {104, 192, 240}}};
-            if(exceptions.at(precision).count(length[0]))
-                intrinsicMode = IntrinsicAccessType::DISABLE_BOTH;
-        }
+    // case 1: is runing tuning or a tuned solution, then use the setting in the config
+    if(specified_key != nullptr)
+    {
+        auto& config  = std::get<4>(*specified_key.get());
+        intrinsicMode = (config.intrinsic_buffer_inst) ? IntrinsicAccessType::ENABLE_BOTH
+                                                       : IntrinsicAccessType::DISABLE_BOTH;
+        return;
+    }
+
+    // case 2: un-tuned: auto decision: try to use buffer instruction as possible
+    if(dir2regMode == TRY_ENABLE_IF_SUPPORT)
+    {
+        if(placement == rocfft_placement_inplace)
+            intrinsicMode = IntrinsicAccessType::ENABLE_LOAD_ONLY;
+        else
+            intrinsicMode = IntrinsicAccessType::ENABLE_BOTH;
+    }
+}
+
+// NB: remember set this value at this point instead of SetupGPAndFnPtr_internal()
+//     since we might need to pass this value to RTC generator
+void SBCCNode::TuneIntrinsicMode()
+{
+    // already disabled
+    if(intrinsicMode == IntrinsicAccessType::DISABLE_BOTH)
+        return;
+
+    // hardcoded turn-off in some exception cases
+    // 1. currently we only enable this on 906, 908, 1030. TODO- test on 90a
+    if((is_device_gcn_arch(deviceProp, "gfx906") == false)
+       && (is_device_gcn_arch(deviceProp, "gfx908") == false)
+       && (is_device_gcn_arch(deviceProp, "gfx1030") == false))
+    {
+        intrinsicMode = IntrinsicAccessType::DISABLE_BOTH;
+    }
+    // 2. exception cases on 906. Based on benchmark results
+    else if(is_device_gcn_arch(deviceProp, "gfx906"))
+    {
+        // bad results from benchmark:
+        // {96,sp}, {125,sp}, {192,sp/dp}, {240,dp}, {256,sp/dp}, {343,sp/dp}
+        std::map<rocfft_precision, std::set<size_t>> exceptions
+            = {{rocfft_precision_single, {96, 125, 192, 256, 343}},
+               {rocfft_precision_double, {192, 240, 256, 343}}};
+        if(length_excepted(exceptions, precision, length[0]))
+            intrinsicMode = IntrinsicAccessType::DISABLE_BOTH;
+    }
+    // 3. exception cases on 908. Based on benchmark results
+    else if(is_device_gcn_arch(deviceProp, "gfx908"))
+    {
+        // bad results from benchmark:
+        // {104,sp/dp}, {192,dp}, {240,dp}, {289,sp}
+        std::map<rocfft_precision, std::set<size_t>> exceptions
+            = {{rocfft_precision_single, {104, 289}}, {rocfft_precision_double, {104, 192, 240}}};
+        if(length_excepted(exceptions, precision, length[0]))
+            intrinsicMode = IntrinsicAccessType::DISABLE_BOTH;
     }
 }
 
 void SBCCNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
 {
-    auto kernel = function_pool::get_kernel(fpkey(length[0], precision, scheme));
+    auto kernel = function_pool::get_kernel(GetKernelKey());
     fnPtr       = kernel.device_function;
     bwd         = kernel.transforms_per_block;
     wgs         = kernel.workgroup_size;
     lds         = length[0] * bwd;
     gp.b_x      = ((length[1]) - 1) / bwd + 1;
     gp.b_x *= std::accumulate(length.begin() + 2, length.end(), batch, std::multiplies<size_t>());
-    gp.wgs_x = kernel.workgroup_size;
+    gp.wgs_x = wgs;
 }
 
 std::vector<size_t> SBCCNode::CollapsibleDims()
 {
+    // do not collapse on multi-kernel fused Bluestein nodes
+    if(typeBlue == BT_MULTI_KERNEL_FUSED)
+        return {};
+
     // second-fastest dim is FFT, higher dims are collapsible
     std::vector<size_t> ret(length.size() - 2);
     std::iota(ret.begin(), ret.end(), 2);
@@ -879,25 +1108,63 @@ std::vector<size_t> SBCCNode::CollapsibleDims()
 /*****************************************************
  * SBRC  *
  *****************************************************/
-bool SBRCNode::KernelCheck()
+FMKey SBRCNode::GetKernelKey() const
 {
-    bool res = LeafNode::KernelCheck();
+    if(specified_key)
+        return *specified_key.get();
 
-    // set according to benchmark
-    SetDirectRegType();
+    // NB: Need to make sure that sbrcTranstype has the correct value
+    if(sbrcTranstype == SBRC_TRANSPOSE_TYPE::NONE)
+    {
+        // find the base kernel at first
+        FMKey baseKey = fpkey(length[0], precision, scheme, TILE_ALIGNED);
+        // if we have the base kernel, then we set the exact sbrc_trans_type and return the real key
+        // if we don't, then we simply return a key with NONE sbrc_trans_type
+        // which will make KernelCheck() trigger an exception
+        if(function_pool::has_function(baseKey))
+        {
+            auto bwd      = function_pool::get_kernel(baseKey).transforms_per_block;
+            sbrcTranstype = sbrc_transpose_type(bwd);
+        }
+    }
 
-    return res;
+    return fpkey(length[0], precision, scheme, sbrcTranstype);
 }
 
-void SBRCNode::SetDirectRegType()
+bool SBRCNode::KernelCheck(std::vector<FMKey>& kernel_keys)
 {
+    bool res = LeafNode::KernelCheck(kernel_keys);
+    if(!res)
+        return false;
+
+    // if we are doing tuning or running with the tuned solution, we have the specified_key.
+    // we must directly run the kernel with the exact setting as the config
+    // without the hardcoded tuning
+    if(specified_key != nullptr)
+        return true;
+
+    // hardocded-tuning according to benchmark
+    TuneDirectRegType();
+
+    return true;
+}
+
+void SBRCNode::TuneDirectRegType()
+{
+    // half precision has not been tested yet, disable it for now.
+    if(precision == rocfft_precision_half)
+    {
+        dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
+        return;
+    }
+
     if(is_device_gcn_arch(deviceProp, "gfx906"))
     {
         // bad results from benchmark:
         // {49,sp}, {128,sp}, {64,dp}, {81,dp}, {100,dp}
         std::map<rocfft_precision, std::set<size_t>> exceptions
             = {{rocfft_precision_single, {49, 128}}, {rocfft_precision_double, {64, 81, 100}}};
-        if(exceptions.at(precision).count(length[0]))
+        if(length_excepted(exceptions, precision, length[0]))
             dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
     }
     else if(is_device_gcn_arch(deviceProp, "gfx908"))
@@ -907,7 +1174,7 @@ void SBRCNode::SetDirectRegType()
         std::map<rocfft_precision, std::set<size_t>> exceptions
             = {{rocfft_precision_single, {81, 100, 128, 192, 200, 512}},
                {rocfft_precision_double, {125, 128}}};
-        if(exceptions.at(precision).count(length[0]))
+        if(length_excepted(exceptions, precision, length[0]))
             dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
     }
     else if(is_device_gcn_arch(deviceProp, "gfx90a"))
@@ -917,7 +1184,7 @@ void SBRCNode::SetDirectRegType()
         std::map<rocfft_precision, std::set<size_t>> exceptions
             = {{rocfft_precision_single, {49, 81, 100, 125, 200, 512}},
                {rocfft_precision_double, {64, 81, 100, 125}}};
-        if(exceptions.at(precision).count(length[0]))
+        if(length_excepted(exceptions, precision, length[0]))
             dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
     }
     // we don't enable the features for others
@@ -929,42 +1196,61 @@ void SBRCNode::SetDirectRegType()
 
 void SBRCNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
 {
-    auto kernel   = function_pool::get_kernel(fpkey(length[0], precision, scheme));
-    bwd           = kernel.transforms_per_block;
-    wgs           = kernel.workgroup_size;
-    lds           = length[0] * bwd;
-    sbrcTranstype = sbrc_transpose_type(bwd);
-    fnPtr         = function_pool::get_function(fpkey(length[0], precision, scheme, sbrcTranstype));
-    gp.b_x        = (length[1] - 1) / bwd + 1;
+    // sbrcTransType has already been assigned in KernelCheck();
+    auto kernel = function_pool::get_kernel(GetKernelKey());
+    fnPtr       = kernel.device_function;
+    bwd         = kernel.transforms_per_block;
+    wgs         = kernel.workgroup_size;
+    lds         = length[0] * bwd;
+    gp.b_x      = (length[1] - 1) / bwd + 1;
     gp.b_x *= std::accumulate(length.begin() + 2, length.end(), batch, std::multiplies<size_t>());
-    gp.wgs_x = kernel.workgroup_size;
+    gp.wgs_x = wgs;
 }
 
 SBRC_TRANSPOSE_TYPE SBRCNode::sbrc_transpose_type(unsigned int blockWidth) const
 {
-    // NB: Since we need a NONE as default, so this NONE is actually TILE_ALIGNED
     auto alignment_dimension = length[1];
-    return (alignment_dimension % blockWidth == 0) ? NONE : TILE_UNALIGNED;
+    return (alignment_dimension % blockWidth == 0) ? TILE_ALIGNED : TILE_UNALIGNED;
 }
 
 /*****************************************************
  * SBCR  *
  *****************************************************/
-bool SBCRNode::KernelCheck()
+bool SBCRNode::KernelCheck(std::vector<FMKey>& kernel_keys)
 {
-    bool res = LeafNode::KernelCheck();
+    bool res = LeafNode::KernelCheck(kernel_keys);
+    if(!res)
+        return false;
 
-    // set according to benchmark
-    SetDirectRegType();
+    // if we are doing tuning or running with the tuned solution, we have the specified_key.
+    // we must directly run the kernel with the exact setting as the config
+    // without the hardcoded tuning
+    if(specified_key != nullptr)
+    {
+        InitIntrinsicMode();
+        return true;
+    }
 
-    // set according to benchmark
-    SetIntrinsicMode();
+    // hardocded-tuning according to benchmark
+    TuneDirectRegType();
 
-    return res;
+    // check if we can use buffer instr
+    InitIntrinsicMode();
+    // hardocded-tuning according to benchmark
+    TuneIntrinsicMode();
+
+    return true;
 }
 
-void SBCRNode::SetDirectRegType()
+void SBCRNode::TuneDirectRegType()
 {
+    // half precision has not been tested yet, disable it for now.
+    if(precision == rocfft_precision_half)
+    {
+        dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
+        return;
+    }
+
     // switch on/off according to the arch
     // tweaking the setting based on the benchmark results.
 
@@ -982,33 +1268,60 @@ void SBCRNode::SetDirectRegType()
     }
 }
 
-void SBCRNode::SetIntrinsicMode()
+void SBCRNode::InitIntrinsicMode()
 {
-    // NB: remember set this value at this point instead of SetupGPAndFnPtr_internal()
-    //     since we might need to pass this value to RTC generator
-    intrinsicMode = IntrinsicAccessType::DISABLE_BOTH;
-
-    // TODO- To test on gfx90a
-    if(is_device_gcn_arch(deviceProp, "gfx908") && (dir2regMode == TRY_ENABLE_IF_SUPPORT))
+    // 1. General rejections: (Guard) cases we definitely can't use buffer instruction
+    // 2. half precision has not been tested yet, disable it for now.
+    if(((uint64_t)iDist * batch * complex_type_size(precision) >= 0xFFFFFFFF)
+       || ((uint64_t)oDist * batch * complex_type_size(precision) >= 0xFFFFFFFF)
+       || (precision == rocfft_precision_half))
     {
-        if(((uint64_t)iDist * batch * sizeof_precision(precision) < 0xFFFFFFFF)
-           && ((uint64_t)oDist * batch * sizeof_precision(precision) < 0xFFFFFFFF))
-        {
-            intrinsicMode = IntrinsicAccessType::ENABLE_BOTH;
-        }
+        intrinsicMode = IntrinsicAccessType::DISABLE_BOTH;
+        return;
+    }
+
+    // case 1: is runing tuning or a tuned solution, then use the setting in the config
+    if(specified_key != nullptr)
+    {
+        auto& config  = std::get<4>(*specified_key.get());
+        intrinsicMode = (config.intrinsic_buffer_inst) ? IntrinsicAccessType::ENABLE_BOTH
+                                                       : IntrinsicAccessType::DISABLE_BOTH;
+        return;
+    }
+
+    // case 2: un-tuned: auto decision: try to use buffer instruction as possible
+    if(dir2regMode == TRY_ENABLE_IF_SUPPORT)
+    {
+        intrinsicMode = IntrinsicAccessType::ENABLE_BOTH;
+    }
+}
+
+// NB: remember set this value at this point instead of SetupGPAndFnPtr_internal()
+//     since we might need to pass this value to RTC generator
+void SBCRNode::TuneIntrinsicMode()
+{
+    // already disabled
+    if(intrinsicMode == IntrinsicAccessType::DISABLE_BOTH)
+        return;
+
+    // hardcoded turn-off in some exception cases
+    // 1. currently we only enable this on 908. TODO- test on 90a
+    if(is_device_gcn_arch(deviceProp, "gfx908") == false)
+    {
+        intrinsicMode = IntrinsicAccessType::DISABLE_BOTH;
     }
 }
 
 void SBCRNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
 {
-    auto kernel = function_pool::get_kernel(fpkey(length[0], precision, scheme));
+    auto kernel = function_pool::get_kernel(GetKernelKey());
     fnPtr       = kernel.device_function;
     wgs         = kernel.workgroup_size;
     bwd         = kernel.transforms_per_block;
     lds         = length[0] * bwd;
     gp.b_x      = ((length[1]) - 1) / bwd + 1;
     gp.b_x *= std::accumulate(length.begin() + 2, length.end(), batch, std::multiplies<size_t>());
-    gp.wgs_x = kernel.workgroup_size;
+    gp.wgs_x = wgs;
 
     if(ebtype != EmbeddedType::NONE)
         lds_padding = 1;
@@ -1016,8 +1329,8 @@ void SBCRNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
     return;
 }
 
-bool SBCRNode::CreateTwiddleTableResource()
+bool SBCRNode::CreateDeviceResources()
 {
     twd_attach_halfN = (ebtype != EmbeddedType::NONE);
-    return LeafNode::CreateTwiddleTableResource();
+    return LeafNode::CreateDeviceResources();
 }
