@@ -1,4 +1,4 @@
-// Copyright (C) 2021 - 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2021 - 2023 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -23,21 +23,28 @@
 #include "function_pool.h"
 #include "logging.h"
 #include "node_factory.h"
+#include "tuning_helper.h"
 #include <numeric>
 
 /*****************************************************
  * 3D_RTRT  *
  *****************************************************/
-void RTRT3DNode::BuildTree_internal()
+void RTRT3DNode::BuildTree_internal(const SchemeVec& child_schemes)
 {
+    bool noSolution = child_schemes.empty();
+
     // 2d fft
     NodeMetaData xyPlanData(this);
     xyPlanData.length    = length;
     xyPlanData.dimension = 2;
-    auto xyPlan          = NodeFactory::CreateExplicitNode(xyPlanData, this);
+    // skip the decide scheme part in node factory
+    ComputeScheme determined_scheme = (noSolution) ? CS_NONE : child_schemes[0];
+    auto          xyPlan = NodeFactory::CreateExplicitNode(xyPlanData, this, determined_scheme);
     xyPlan->RecursiveBuildTree();
 
     // first transpose
+    if(!noSolution)
+        assert(child_schemes[1] == CS_KERNEL_TRANSPOSE_XY_Z);
     auto trans1Plan    = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE_XY_Z, this);
     trans1Plan->length = length;
     trans1Plan->SetTransposeOutputLength();
@@ -50,10 +57,14 @@ void RTRT3DNode::BuildTree_internal()
     zPlanData.length.push_back(length[2]);
     zPlanData.length.push_back(length[0]);
     zPlanData.length.push_back(length[1]);
-    auto zPlan = NodeFactory::CreateExplicitNode(zPlanData, this);
+    // skip the decide scheme part in node factory
+    determined_scheme = determined_scheme = (noSolution) ? CS_NONE : child_schemes[2];
+    auto zPlan = NodeFactory::CreateExplicitNode(zPlanData, this, determined_scheme);
     zPlan->RecursiveBuildTree();
 
     // second transpose
+    if(!noSolution)
+        assert(child_schemes[3] == CS_KERNEL_TRANSPOSE_Z_XY);
     auto trans2Plan    = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE_Z_XY, this);
     trans2Plan->length = zPlan->length;
     trans2Plan->SetTransposeOutputLength();
@@ -143,13 +154,17 @@ void RTRT3DNode::AssignParams_internal()
 /*****************************************************
  * 3D_TRTRTR  *
  *****************************************************/
-void TRTRTR3DNode::BuildTree_internal()
+void TRTRTR3DNode::BuildTree_internal(const SchemeVec& child_schemes)
 {
+    bool noSolution = child_schemes.empty();
+
     std::vector<size_t> cur_length = length;
 
     for(int i = 0; i < 6; i += 2)
     {
         // transpose Z_XY
+        if(!noSolution)
+            assert(child_schemes[i] == CS_KERNEL_TRANSPOSE_Z_XY);
         auto trans_plan    = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE_Z_XY, this);
         trans_plan->length = cur_length;
         trans_plan->SetTransposeOutputLength();
@@ -162,7 +177,9 @@ void TRTRTR3DNode::BuildTree_internal()
         NodeMetaData row_plan_data(this);
         row_plan_data.length    = cur_length;
         row_plan_data.dimension = 1;
-        auto row_plan           = NodeFactory::CreateExplicitNode(row_plan_data, this);
+        // skip the decide scheme part in node factory
+        ComputeScheme determined_scheme = (noSolution) ? CS_NONE : child_schemes[i + 1];
+        auto row_plan = NodeFactory::CreateExplicitNode(row_plan_data, this, determined_scheme);
         row_plan->RecursiveBuildTree();
 
         // TR
@@ -264,8 +281,10 @@ void TRTRTR3DNode::AssignParams_internal()
 /*****************************************************
  * CS_3D_BLOCK_RC  *
  *****************************************************/
-void BLOCKRC3DNode::BuildTree_internal()
+void BLOCKRC3DNode::BuildTree_internal(const SchemeVec& child_schemes)
 {
+    bool noSolution = child_schemes.empty();
+
     std::vector<size_t> cur_length = length;
 
     // NB:
@@ -317,7 +336,7 @@ void BLOCKRC3DNode::BuildTree_internal()
         if(have_sbrc)
         {
             auto kernel = function_pool::get_kernel(
-                fpkey(cur_length[0], precision, CS_KERNEL_STOCKHAM_BLOCK_RC));
+                fpkey(cur_length[0], precision, CS_KERNEL_STOCKHAM_BLOCK_RC, TILE_ALIGNED));
 
             size_t otherDim = use_ZXY_sbrc ? cur_length[1] : cur_length[2];
             if(otherDim % kernel.transforms_per_block != 0)
@@ -344,8 +363,10 @@ void BLOCKRC3DNode::BuildTree_internal()
 
         if(have_sbrc)
         {
-            auto sbrcScheme   = (use_ZXY_sbrc) ? CS_KERNEL_STOCKHAM_TRANSPOSE_Z_XY
-                                               : CS_KERNEL_STOCKHAM_TRANSPOSE_XY_Z;
+            auto sbrcScheme = (use_ZXY_sbrc) ? CS_KERNEL_STOCKHAM_TRANSPOSE_Z_XY
+                                             : CS_KERNEL_STOCKHAM_TRANSPOSE_XY_Z;
+            if(!noSolution)
+                assert(child_schemes[childNodes.size()] == sbrcScheme);
             auto sbrc_node    = NodeFactory::CreateNodeFromScheme(sbrcScheme, this);
             sbrc_node->length = cur_length;
             sbrc_node->SetTransposeOutputLength();
@@ -357,12 +378,16 @@ void BLOCKRC3DNode::BuildTree_internal()
             NodeMetaData row_plan_data(this);
             row_plan_data.length    = cur_length;
             row_plan_data.dimension = 1;
-            auto row_plan           = NodeFactory::CreateExplicitNode(row_plan_data, this);
+            ComputeScheme determined_scheme
+                = (noSolution) ? CS_NONE : child_schemes[childNodes.size()];
+            auto row_plan = NodeFactory::CreateExplicitNode(row_plan_data, this, determined_scheme);
             row_plan->RecursiveBuildTree();
 
             // transpose XY_Z
             auto transScheme = (use_ZXY_sbrc) ? CS_KERNEL_TRANSPOSE_Z_XY : CS_KERNEL_TRANSPOSE_XY_Z;
-            auto trans_plan  = NodeFactory::CreateNodeFromScheme(transScheme, this);
+            if(!noSolution)
+                assert(child_schemes[childNodes.size() + 1] == transScheme);
+            auto trans_plan    = NodeFactory::CreateNodeFromScheme(transScheme, this);
             trans_plan->length = cur_length;
             trans_plan->SetTransposeOutputLength();
             if(!use_ZXY_sbrc)
@@ -389,11 +414,6 @@ void BLOCKRC3DNode::BuildTree_internal()
 
 void BLOCKRC3DNode::AssignParams_internal()
 {
-    // could go as low as 3 kernels if all dimensions are SBRC-able,
-    // but less than 6.  If we ended up with 6 we should have just
-    // done 3D_TRTRTR instead.
-    assert(childNodes.size() >= 3 && childNodes.size() < 6);
-
     childNodes.front()->inStride = inStride;
     childNodes.front()->iDist    = iDist;
 
@@ -474,14 +494,18 @@ void BLOCKRC3DNode::AssignParams_internal()
 /*****************************************************
  * CS_3D_BLOCK_CR  *
  *****************************************************/
-void BLOCKCR3DNode::BuildTree_internal()
+void BLOCKCR3DNode::BuildTree_internal(const SchemeVec& child_schemes)
 {
+    bool noSolution = child_schemes.empty();
+
     // TODO: It works only for 3 SBCR children nodes for now.
     //       The final logic will be similar to what SBRC has.
 
     std::vector<size_t> cur_length = length;
     for(int i = 0; i < 3; ++i)
     {
+        if(!noSolution)
+            assert(child_schemes[i] == CS_KERNEL_STOCKHAM_BLOCK_CR);
         auto node = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CR, this);
         node->length.push_back(cur_length[2]);
         node->length.push_back(cur_length[0] * cur_length[1]);
@@ -526,8 +550,10 @@ void BLOCKCR3DNode::AssignParams_internal()
 /*****************************************************
  * CS_3D_RC  *
  *****************************************************/
-void RC3DNode::BuildTree_internal()
+void RC3DNode::BuildTree_internal(const SchemeVec& child_schemes)
 {
+    bool noSolution = child_schemes.empty();
+
     // 2d fft
     NodeMetaData xyPlanData(this);
     xyPlanData.length.push_back(length[0]);
@@ -538,7 +564,9 @@ void RC3DNode::BuildTree_internal()
     {
         xyPlanData.length.push_back(length[index]);
     }
-    auto xyPlan = NodeFactory::CreateExplicitNode(xyPlanData, this);
+    // skip the decide scheme part in node factory
+    ComputeScheme determined_scheme = (noSolution) ? CS_NONE : child_schemes[0];
+    auto          xyPlan = NodeFactory::CreateExplicitNode(xyPlanData, this, determined_scheme);
     xyPlan->RecursiveBuildTree();
 
     // z col fft
@@ -555,16 +583,27 @@ void RC3DNode::BuildTree_internal()
 
     // use explicit SBCC kernel if available
     std::unique_ptr<TreeNode> zPlan;
-    if(function_pool::has_SBCC_kernel(length[2], precision))
+
+    // skip the decide scheme part in node factory
+    determined_scheme = (noSolution) ? CS_NONE : child_schemes[1];
+    if(determined_scheme)
     {
-        zPlan            = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CC, this);
-        zPlan->length    = zPlanData.length;
-        zPlan->dimension = 1;
+        zPlan = NodeFactory::CreateExplicitNode(zPlanData, this, determined_scheme);
+        zPlan->RecursiveBuildTree();
     }
     else
     {
-        zPlan = NodeFactory::CreateExplicitNode(zPlanData, this);
-        zPlan->RecursiveBuildTree();
+        if(function_pool::has_SBCC_kernel(length[2], precision))
+        {
+            zPlan            = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CC, this);
+            zPlan->length    = zPlanData.length;
+            zPlan->dimension = 1;
+        }
+        else
+        {
+            zPlan = NodeFactory::CreateExplicitNode(zPlanData, this);
+            zPlan->RecursiveBuildTree();
+        }
     }
 
     // RC
@@ -602,42 +641,56 @@ void RC3DNode::AssignParams_internal()
 /*****************************************************
  * Base Class of fused SBRC and Transpose
  *****************************************************/
-bool SBRCTranspose3DNode::KernelCheck()
+FMKey SBRCTranspose3DNode::GetKernelKey() const
 {
-    // check we have the kernel,
-    // we always have aligned, get the kernel and the bwd
-    FMKey key = fpkey(length[0], precision, scheme, TILE_ALIGNED);
-    if(!function_pool::has_function(key))
+    if(specified_key)
+        return *specified_key.get();
+
+    // NB: Need to make sure that sbrcTranstype has the correct value
+    if(sbrcTranstype == SBRC_TRANSPOSE_TYPE::NONE)
     {
-        if(LOG_TRACE_ENABLED())
-            (*LogSingleton::GetInstance().GetTraceOS()) << PrintMissingKernelInfo(key);
-        return false;
+        // find the base kernel at first
+        FMKey baseKey = fpkey(length[0], precision, scheme, TILE_ALIGNED);
+        // if we have the base kernel, then we set the exact sbrc_trans_type and return the real key
+        // if we don't, then we simply return a key with NONE sbrc_trans_type
+        // which will make KernelCheck() trigger an exception
+        if(function_pool::has_function(baseKey))
+        {
+            auto bwd      = function_pool::get_kernel(baseKey).transforms_per_block;
+            sbrcTranstype = sbrc_transpose_type(bwd);
+        }
     }
 
-    auto bwd = function_pool::get_kernel(key).transforms_per_block;
+    return fpkey(length[0], precision, scheme, sbrcTranstype);
+}
 
-    // check if we have the sbrc_type that we are actually applying
-    sbrcTranstype = sbrc_transpose_type(bwd);
-    if(!function_pool::has_function(fpkey(length[0], precision, scheme, sbrcTranstype)))
-    {
-        if(LOG_TRACE_ENABLED())
-            (*LogSingleton::GetInstance().GetTraceOS())
-                << PrintMissingKernelInfo(fpkey(length[0], precision, scheme, sbrcTranstype));
+bool SBRCTranspose3DNode::KernelCheck(std::vector<FMKey>& kernel_keys)
+{
+    bool res = LeafNode::KernelCheck(kernel_keys);
+    if(!res)
         return false;
-    }
 
-    dir2regMode = (function_pool::get_kernel(key).direct_to_from_reg)
-                      ? DirectRegType::TRY_ENABLE_IF_SUPPORT
-                      : DirectRegType::FORCE_OFF_OR_NOT_SUPPORT;
+    // if we are doing tuning or running with the tuned solution, we have the specified_key.
+    // we must directly run the kernel with the exact setting as the config
+    // without the hardcoded tuning
+    if(specified_key != nullptr)
+        return true;
 
-    // set according to benchmark
-    SetDirectRegType();
+    // hardocded-tuning according to benchmark
+    TuneDirectRegType();
 
     return true;
 }
 
-void SBRCTranspose3DNode::SetDirectRegType()
+void SBRCTranspose3DNode::TuneDirectRegType()
 {
+    // half precision has not been tested yet, disable it for now.
+    if(precision == rocfft_precision_half)
+    {
+        dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
+        return;
+    }
+
     if(is_device_gcn_arch(deviceProp, "gfx906"))
     {
         // bad results from benchmark:
@@ -645,7 +698,7 @@ void SBRCTranspose3DNode::SetDirectRegType()
         //     {64,dp}, {81,dp}, {100,dp} are bad
         std::map<rocfft_precision, std::set<size_t>> exceptions
             = {{rocfft_precision_single, {49, 128}}, {rocfft_precision_double, {64, 81, 100}}};
-        if(exceptions.at(precision).count(length[0]))
+        if(length_excepted(exceptions, precision, length[0]))
             dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
     }
     else if(is_device_gcn_arch(deviceProp, "gfx908"))
@@ -656,7 +709,7 @@ void SBRCTranspose3DNode::SetDirectRegType()
         std::map<rocfft_precision, std::set<size_t>> exceptions
             = {{rocfft_precision_single, {81, 100, 128, 192, 200, 512}},
                {rocfft_precision_double, {81, 512}}};
-        if(exceptions.at(precision).count(length[0]))
+        if(length_excepted(exceptions, precision, length[0]))
             dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
     }
     else if(is_device_gcn_arch(deviceProp, "gfx90a"))
@@ -667,7 +720,7 @@ void SBRCTranspose3DNode::SetDirectRegType()
         std::map<rocfft_precision, std::set<size_t>> exceptions
             = {{rocfft_precision_single, {49, 64, 81, 125, 128, 192, 200, 512}},
                {rocfft_precision_double, {64, 81, 100, 125}}};
-        if(exceptions.at(precision).count(length[0]))
+        if(length_excepted(exceptions, precision, length[0]))
             dir2regMode = FORCE_OFF_OR_NOT_SUPPORT;
     }
     // we don't enable the features for others
@@ -683,15 +736,14 @@ void SBRCTranspose3DNode::SetDirectRegType()
  *****************************************************/
 void SBRCTransXY_ZNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
 {
-    auto kernel
-        = function_pool::get_kernel(fpkey(length[0], precision, CS_KERNEL_STOCKHAM_BLOCK_RC));
-    bwd           = kernel.transforms_per_block;
-    wgs           = kernel.workgroup_size;
-    lds           = length[0] * bwd;
-    sbrcTranstype = sbrc_transpose_type(bwd);
-    fnPtr         = function_pool::get_function(fpkey(length[0], precision, scheme, sbrcTranstype));
-    gp.b_x        = DivRoundingUp(length[2], bwd) * length[1] * batch;
-    gp.wgs_x      = kernel.workgroup_size;
+    // sbrcTransType has already been assigned in KernelCheck();
+    auto kernel = function_pool::get_kernel(GetKernelKey());
+    fnPtr       = kernel.device_function;
+    bwd         = kernel.transforms_per_block;
+    wgs         = kernel.workgroup_size;
+    lds         = length[0] * bwd;
+    gp.b_x      = DivRoundingUp(length[2], bwd) * length[1] * batch;
+    gp.wgs_x    = wgs;
 }
 
 /*****************************************************
@@ -700,15 +752,14 @@ void SBRCTransXY_ZNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp
  *****************************************************/
 void SBRCTransZ_XYNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
 {
-    auto kernel
-        = function_pool::get_kernel(fpkey(length[0], precision, CS_KERNEL_STOCKHAM_BLOCK_RC));
-    bwd           = kernel.transforms_per_block;
-    wgs           = kernel.workgroup_size;
-    lds           = length[0] * bwd;
-    sbrcTranstype = sbrc_transpose_type(bwd);
-    fnPtr         = function_pool::get_function(fpkey(length[0], precision, scheme, sbrcTranstype));
-    gp.b_x        = DivRoundingUp(length[1], bwd) * length[2] * batch;
-    gp.wgs_x      = kernel.workgroup_size;
+    // sbrcTransType has already been assigned in KernelCheck();
+    auto kernel = function_pool::get_kernel(GetKernelKey());
+    fnPtr       = kernel.device_function;
+    bwd         = kernel.transforms_per_block;
+    wgs         = kernel.workgroup_size;
+    lds         = length[0] * bwd;
+    gp.b_x      = DivRoundingUp(length[1], bwd) * length[2] * batch;
+    gp.wgs_x    = wgs;
 }
 
 /*****************************************************
@@ -717,16 +768,15 @@ void SBRCTransZ_XYNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp
  *****************************************************/
 void RealCmplxTransZ_XYNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
 {
-    auto kernel
-        = function_pool::get_kernel(fpkey(length[0], precision, CS_KERNEL_STOCKHAM_BLOCK_RC));
-    bwd           = kernel.transforms_per_block;
-    wgs           = kernel.workgroup_size;
-    lds           = length[0] * bwd;
-    lds_padding   = 1;
-    sbrcTranstype = sbrc_transpose_type(bwd);
-    fnPtr         = function_pool::get_function(fpkey(length[0], precision, scheme, sbrcTranstype));
-    gp.b_x        = DivRoundingUp(length[1], bwd) * length[2] * batch;
-    gp.wgs_x      = kernel.workgroup_size;
+    // sbrcTransType has already been assigned in KernelCheck();
+    auto kernel = function_pool::get_kernel(GetKernelKey());
+    fnPtr       = kernel.device_function;
+    bwd         = kernel.transforms_per_block;
+    wgs         = kernel.workgroup_size;
+    lds_padding = 1;
+    lds         = (length[0] + lds_padding) * bwd;
+    gp.b_x      = DivRoundingUp(length[1], bwd) * length[2] * batch;
+    gp.wgs_x    = wgs;
 }
 
 bool RealCmplxTransZ_XYNode::CreateDevKernelArgs()
