@@ -25,10 +25,14 @@ from sympy import sieve
 
 import numpy as np
 
+import perflib.specs
+
 all_precisions = ['single', 'double']
 all_directions = [-1, 1]
 all_inplaces = [True, False]
 all_reals = [True, False]
+def_tuning_min_wgs = 64
+def_tuning_max_wgs = 512
 
 # yapf: disable
 lengths = {
@@ -189,9 +193,9 @@ lengths = {
     ],
 
     'qa2d10b': [
+        (3125, 3125),
         (4096, 4096),
         (6561, 6561),
-        (3125, 3125),
     ],
 
     'qa3d10b': [
@@ -343,12 +347,25 @@ def mktag(tag, dimension, precision, direction, inplace, real):
 
 # yield problem sizes with default precision, direction, etc
 def default_length_params(tag, lengths, nbatch, precisions=all_precisions, \
-    directions=all_directions, inplaces=all_inplaces, reals=all_reals):
+    directions=all_directions, inplaces=all_inplaces, reals=all_reals, min_wgs=def_tuning_min_wgs, max_wgs=def_tuning_max_wgs):
+
+    # workaround: disable failing token on gfx906
+    if perflib.specs.get_machine_specs(0).gpuid == '0x66a1':
+        gfx906 = True
+    else:
+        gfx906 = False
 
     for precision, direction, inplace, real in product(precisions, directions,
                                                        inplaces, reals):
         for length in lengths:
             length = (length, ) if isinstance(length, int) else length
+
+            # workaround: disable failing token on gfx906
+            if gfx906 and (length == [32768, 32768] and nbatch == 1
+                           and direction == -1 and not inplace and real
+                           and precision == 'single'):
+                continue
+
             yield Problem(length,
                           tag=mktag(tag, len(length), precision, direction,
                                     inplace, real),
@@ -356,7 +373,9 @@ def default_length_params(tag, lengths, nbatch, precisions=all_precisions, \
                           direction=direction,
                           inplace=inplace,
                           real=real,
-                          precision=precision)
+                          precision=precision,
+                          min_wgs=min_wgs,
+                          max_wgs=max_wgs)
 
 
 def md():
@@ -369,7 +388,7 @@ def qa():
     """AMD QA suite."""
 
     for length1 in [
-            8192, 10752, 18816, 21504, 32256, 43008, 16384, 19683, 15625, 16807
+            8192, 10752, 15625, 16384, 16807, 18816, 19683, 21504, 32256, 43008
     ]:
         for direction in [-1, 1]:
             yield Problem([length1],
@@ -389,13 +408,12 @@ def qa():
                   real=False,
                   precision='double')
 
-    yield Problem((336, 336, 56),
-                  tag=mktag('qa3', 3, 'double', -1, False, False),
-                  nbatch=1,
-                  direction=-1,
-                  inplace=False,
-                  real=False,
-                  precision='double')
+    yield from default_length_params("336x336x56", [(336, 336, 56)],
+                                     1,
+                                     directions=[-1],
+                                     precisions=['double'],
+                                     inplaces=[True, False],
+                                     reals=[False])
 
     for length3 in lengths['md']:
         for direction in [-1, 1]:
@@ -406,7 +424,8 @@ def qa():
                           direction=direction,
                           inplace=False,
                           real=True,
-                          precision='single')
+                          precision='single',
+                          meta={'figtype': 'bargraph'})
 
     for length in lengths['qa1d10b']:
         yield Problem([length],
@@ -630,6 +649,27 @@ def batched_1d_small_r2c():
                                      reals=[True])
 
 
+def batch_const_count():
+    # batch * length = 2^25 ... 2^30
+    for direction in [-1, 1]:
+        for precision in all_precisions:
+            for exp in [25, 26, 27, 28, 29, 30]:
+                for place in all_inplaces:
+                    for lexp in range(4, exp + 1):
+                        length = 2**lexp
+                        batch = 2**(exp - lexp)
+
+                        yield Problem([length],
+                                      tag=mktag("footprint2exp" + str(exp), 1,
+                                                precision, direction, False,
+                                                False),
+                                      nbatch=batch,
+                                      direction=direction,
+                                      inplace=place,
+                                      real=False,
+                                      precision=precision)
+
+
 def benchmarks():
     """Benchmarks: XXX"""
 
@@ -676,3 +716,114 @@ def short_test():
                                                     (4294967296)],
                                      1,
                                      reals=[False])
+
+
+def tuning_example():
+    """tuning 3 examples problems"""
+
+    yield from default_length_params("81_1d", [(81)],
+                                     60000,
+                                     directions=[-1],
+                                     precisions=['double'],
+                                     inplaces=[False],
+                                     reals=[False],
+                                     min_wgs=128,
+                                     max_wgs=256)
+
+    yield from default_length_params("81_2d", [(81, 81)],
+                                     8000,
+                                     directions=[-1],
+                                     precisions=['double'],
+                                     inplaces=[False],
+                                     reals=[False],
+                                     min_wgs=128,
+                                     max_wgs=256)
+
+    # batch=500 to enabling tuning with intrinsic buffer
+    yield from default_length_params("81_3d", [(81, 81, 81)],
+                                     500,
+                                     directions=[-1],
+                                     precisions=['double'],
+                                     inplaces=[False],
+                                     reals=[False],
+                                     min_wgs=128,
+                                     max_wgs=256)
+
+
+def tuning_suite():
+    """tuning"""
+
+    # basically, when tuning for single, we can tune wgs range from 128~512, for double, wgs range is 128~256
+    # but you can also change it for particular problem.
+    # But inside our cpp tuner implementation, min_wgs might still be automatically changed
+    # if the setting gives no any candidate (example, a len64 with min_wgs=128 might not derive any)
+
+    # complex transforms in suite qa.
+    for length1 in [
+            8192, 10000, 10752, 15625, 16384, 16807, 18816, 19683, 21504
+    ]:
+        for direction in [1]:
+            yield Problem([length1],
+                          tag=mktag("qa1", 1, 'double', direction, False,
+                                    False),
+                          nbatch=10000,
+                          direction=direction,
+                          inplace=False,
+                          real=False,
+                          precision='double',
+                          min_wgs=128,
+                          max_wgs=256)
+
+    # batch=5000 to enabling tuning with intrinsic buffer
+    # since batch 10000 causes memory offset > 2^32, buffer inst will be disabled
+    for length1 in [32256, 43008]:
+        for direction in [1]:
+            yield Problem([length1],
+                          tag=mktag("qa1", 1, 'double', direction, False,
+                                    False),
+                          nbatch=5000,
+                          direction=direction,
+                          inplace=False,
+                          real=False,
+                          precision='double',
+                          min_wgs=128,
+                          max_wgs=256)
+
+    # we'd like to search more for this problem, so min_wgs = 64, not 128
+    yield from default_length_params("336x336x56", [(336, 336, 56)],
+                                     1,
+                                     directions=[-1],
+                                     precisions=['double'],
+                                     inplaces=[True, False],
+                                     reals=[False],
+                                     max_wgs=256)
+
+    for length in lengths['qa1d10b']:
+        yield Problem([length],
+                      tag=mktag("qa1d10b", 1, 'single', -1, True, False),
+                      nbatch=10,
+                      direction=-1,
+                      inplace=True,
+                      real=False,
+                      precision='single',
+                      min_wgs=128)
+
+    for length2 in lengths['qa2d10b']:
+        yield Problem(length2,
+                      tag=mktag("qa2d10b", 2, 'single', -1, True, False),
+                      nbatch=10,
+                      direction=-1,
+                      inplace=True,
+                      real=False,
+                      precision='single',
+                      min_wgs=128)
+
+    for length3 in lengths['qa3d10b']:
+        yield Problem(length3,
+                      tag=mktag("qa3d10b", 3, 'single', -1, True, False),
+                      nbatch=10,
+                      direction=-1,
+                      inplace=True,
+                      real=False,
+                      precision='single',
+                      min_wgs=128)
